/* Cable.java
 *
 * Jim McBeath, March 12, 2003
 */

package net.jimmc.spelsim;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/** The cable and attached masses. */
public class Cable {
    private static final int STRESS_COLOR_COUNT = 100;
    protected static final double FWD_CIRCLE_FRACTION = 1e10;

    private boolean debugOutputEnabled;

    /** Total length of the cable in meters.
     * This is the initial length when under the stress applied by the
     * defined capacity.
     */
    double length;

    /** Capacity of the cable, which is how much force can be applied
     * to it at the bottom end before it yields.
     * If the breaking strength of the material is equal to the yield
     * strength, then the cable breaks when more than this much load
     * is applied.
     * If the breaking strength is higher than the yield strength, then
     * the cable becomes inelastic, so remains stretched, breaking when
     * it reaches its breaking strength.
     * Units are Newtons.
     */
    double capacity;

    /** The safety factor.
     * We de-rate the material strength by this amount when calculating
     * how much we need to support our desired capacity.
     */
    double safetyFactor;

    /** The material from which this cable is made. */
    Material material;

    /** The planet around which this cable orbits. */
    Planet planet;

    /** Taper factor from planet end to syncLength point. */
    double taper;

    /** The number of elements to use when modeling the cable in pieces. */
    int numElements = 100;

    //int numBrokenElements = 0;  //the number of broken elements
    int numGroundedElements = 0;  //the number of grounded elements

    /** The elements of the cable for use when modeling in pieces. */
    CableElement[] elements;

    /** The index of the thickest element in the cable. */
    int indexThickest;

    /** The total mass of the cable. */
    double mass;

    /** The required counterweight mass. */
    double cwMass;

    /** The altitude of the center of mass. */
    double centerOfMass;

    /** The altitude of the center of gravity. */
    double centerOfGravity;

    /** The altitude of the center of centrifugal. */
    double centerOfCentrifugal;

    /** A set of known cable values. */
    static Map knownCables;

    private Color[] stressColorMap;

    /** Get a cable with preset values.
     */
    public static Cable getCable(String name) {
        if (knownCables==null) {
	    initKnownCables();
	}
	return (Cable)knownCables.get(name);	//null if not found
    }

    /** Get the list of known cable names. */
    public static String[] getCableNames() {
        if (knownCables==null) {
	    initKnownCables();
	}
        String[] names = (String[])knownCables.keySet().toArray(new String[0]);
	Arrays.sort(names);
	return names;
    }

    private static void initKnownCables() {
	knownCables = new HashMap();
        addCable("Edwards 2000",100e6,20e3,2,100);
        addCable("Pearson 1975",144e6,20e3,2,100);
    }

    private static void addCable(String name, double length,
    		double capacity, double safetyFactor, int numElements) {
        Cable c = new Cable();
	c.setLength(length);
	c.setCapacity(capacity);
	c.setSafetyFactor(safetyFactor);
        c.setNumElements(numElements);
	knownCables.put(name,c);
    }

    /** Create a cable. */
    public Cable() {}

    public void setDebugOutputEnabled(boolean enabled) {
        debugOutputEnabled = enabled;
    }

    public boolean isDebugOutputEnabled() {
        return debugOutputEnabled;
    }

    /** Set the length of the cable.
     * @param length The length of the cable in meters.
     */
    public void setLength(double length) {
    	this.length = length;
	elements = null;	//force recalculation later
    }

    /** Get the length of the cable.
     * @return The length of the cable in meters.
     */
    public double getLength() {
    	return length;
    }

    /** Set the capacity of the cable.
     * @param capacity The capacity of the cable in Newtons.
     */
    public void setCapacity(double capacity) {
    	this.capacity = capacity;
	elements = null;	//force recalculation later
    }

    /** Get the capacity of the cable.
     * @return The capacity of the cable in Newtons.
     */
    public double getCapacity() {
    	return capacity;
    }

    /** Set the safety factor of the cable.
     * @param safetyFactor The safety factor of the cable.
     */
    public void setSafetyFactor(double safetyFactor) {
    	this.safetyFactor = safetyFactor;
	elements = null;	//force recalculation later
    }

    /** Get the safety factor of the cable.
     * @return The safety factor of the cable.
     */
    public double getSafetyFactor() {
    	return safetyFactor;
    }

    /** Set the material of the cable.
     * @param material The material from which the cable is made.
     */
    public void setMaterial(Material material) {
    	this.material = material;
	elements = null;	//force recalculation later
    }

    /** Get the material of the cable.
     * @return The material from which the cable is made.
     */
    public Material getMaterial() {
    	return material;
    }

    /** Set the number of elements in the cable calculations.
     * @param numElements The number of elements to use in
     *          calculations and simulation of the cable.
     */
    public void setNumElements(int numElements) {
    	this.numElements = numElements;
	elements = null;	//force recalculation later
    }

    /** Get the number of elements used for the cable calculations
     *          and simulation.
     * @return The number of elements used for the cable.
     */
    public int getNumElements() {
    	return numElements;
    }

    public int getNumBrokenElements() {
        //return numBrokenElements;
        int numBroken = 0;
        for (int i=0; i<elements.length; i++) {
            if (elements[i].isBroken())
                numBroken++;
        }
        return numBroken;
    }

    public int getNumGroundedElements() {
        int g = 0;
    	for (int i=0; i<numElements; i++) {
            if (elements[i].dyn[0].grounded)
                g++;
        }
        /*
        if (g!=numGroundedElements) {
System.out.println("g="+g+", numGroundedElements="+numGroundedElements);
        }
        */
        return g;
        //return numGroundedElements;
    }

    //Mark the specified element as broken
    public void setElementBroken(int elementNumber) {
        if (elementNumber<0 || elementNumber>=elements.length)
            return;             //ignore out of range index
        elements[elementNumber].setBroken();
    }

    /** Get the current list of elements. */
    public CableElement[] getElements() {
        return elements;
    }

    /** Get the bounding box containing all of the cable elements. */
    public Rectangle2D getBoundingBox() {
        Rectangle2D bbox = new Rectangle2D.Double();
        for (int i=0; i<elements.length; i++) {
            bbox.add(elements[i].dyn[0].d.x,elements[i].dyn[0].d.y);
        }
        return bbox;
    }

    /** Get the maximum radius value for all cable elements. */
    public double getBoundingRadius() {
        double r = 0.0;
        for (int i=0; i<elements.length; i++) {
            double er = elements[i].dyn[0].d.getRadius();
            if (er>r)
                r = er;
        }
        return r;
    }

    /** Set the planet around which the cable orbits.
     * @param planet The planet around which the cable orbits.
     */
    public void setPlanet(Planet planet) {
    	this.planet = planet;
	elements = null;	//force recalculation later
    }

    /** Get the planet aound which the cable orbits.
     * @return The planet aound which the cable orbits.
     */
    public Planet getPlanet() {
    	return planet;
    }

    /** Get the base area of the cable. */
    public double getBaseArea() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return elements[0].area;
    }


    /** Get the total mass of the cable. */
    public double getMass() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return mass;
    }

    /** Get the mass of the counterweight. */
    public double getCounterweightMass() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return cwMass;
    }

    /** Get the cable taper factor.
     * The thickest part of the cable is at synchronous orbit.
     * The taper is the ratio of that thickness to the thickness at the
     * base of the cable.
     */
    public double getTaper() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return elements[indexThickest].area/elements[0].area;
    }

    /** Get the altitude of the center of mass of the cable.
     * @return The altitude of the center of mass over the planet surface
     *         in meters.
     */
    public double getCenterOfMassAltitude() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return centerOfMass;
    }

    /** Get the altitude of the center of gravity of the cable.
     * @return The altitude of the center of gravity over the planet surface
     *         in meters.
     */
    public double getCenterOfGravityAltitude() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return centerOfGravity;
    }

    /** Get the altitude of the center of centrifugal of the cable.
     * @return The altitude of the center of centrifugal over the planet surface
     *         in meters.
     */
    public double getCenterOfCentrifugalAltitude() {
        if (elements==null) {
	    recalculateElements();
	    if (elements==null)
	        return 0.0;		//not enough info
	}
	return centerOfCentrifugal;
    }

    /** Recalculate the elements. */
    public void recalculateElements() {
	elements = null;
        //numBrokenElements = 0;
        numGroundedElements = 0;
        stressColorMap = null;

	//Make sure we have the information we need to do our calculations
        if (material==null || planet==null)
	    return;	//can't do the calculation
	double strength = material.getYieldStrength();
	if (strength==0)
	    return;	//no strength set to do the calculation
	double safetyFactor = getSafetyFactor();
	if (safetyFactor==0)
	    return;
	double stress = strength/safetyFactor;
	double period = planet.getPeriod();
	if (period==0)
	    return;	//can't be done if there is no rotation
	double length = getLength();
	if (length==0)
	    return;

	//precalculate values used in the loop
	double radius = planet.getRadius();
	double elementLength = length/numElements;
	double baseHeight = radius+elementLength;
		//use the top of each element as its height
	double planetMass = planet.getMass();
        double load = getCapacity()*planet.getGravity();  //load in Newtons
	double gMass = Physics.G*planetMass;
	double radial = 2*Math.PI/period;  //angular velocity in radians/sec
	double radial2 = radial*radial;
	double density = material.getDensity();

	//Reset variables collected during element calculation
	mass = 0.0;
	cwMass = 0.0;
	indexThickest = 0;
	centerOfMass = 0;
	centerOfGravity = 0;
	centerOfCentrifugal = 0;
	double sumOfMass = 0;
	double sumOfGravity = 0;
	double sumOfCentrifugal = 0;

	//Create the cable elements.
	//We divide up the cable into numElements elements, all the same
	//length.
	//We design the taper on the cable such that every element starts
	//with the same stress, thus the same strain, thus the unloaded
	//length of every element is also the same.
        elements = new CableElement[numElements];
	double youngs = material.getYoungsModulus();
	double strain = (youngs==0.0)?0.0:(stress/youngs);
	for (int i=0; i<numElements; i++) {
	    CableElement e = new CableElement(this);
	    e.index = i;
	    elements[i] = e;
	    e.area = load/stress;	//m^2
	    e.dyn[0].stress = stress;	//all elements start with same stress
	    e.unloadedLength = elementLength/(1.0+strain);
	    e.Ld = e.unloadedLength*density; //used in acceleration calculations
	    double elementVolume = e.area*e.unloadedLength;	//m^3
	    e.elementMass = elementVolume*density;	//kg
	    mass += e.elementMass;
	    double elementHeight = baseHeight+i*elementLength;
	    double elementGravity = gMass/(elementHeight*elementHeight);
	    double elementCentrifugal = radial2*elementHeight;
	    double elementAcceleration = elementGravity - elementCentrifugal;
	    double elementLoad = e.elementMass*elementAcceleration;
            e.dyn[0].d = new DVector(elementHeight, 0.0);
                //Cable starts off pointing straight along x-axis
            e.dyn[0].v = new DVector(0.0, radial*elementHeight);
                //Cable starts off in synchronized rotation.
	    	//Given that we are looking down at the north pole, and the
		//planet rotates to the east, like Earth, then the direction
		//of the rotation is counterclockwise.  Since our cable
		//starts out pointing right on the screen,
		//it is moving up.
            e.r0 = elementHeight;
	    load += elementLoad;
	    if (load<0.0)
	        load = 0.0;	//that's all the cable we need
	    if (i>0 && e.area>elements[i-1].area)
	        indexThickest = i;
	    centerOfMass += e.elementMass*elementHeight;
	    centerOfGravity += e.elementMass*elementGravity*elementHeight;
	    sumOfGravity += e.elementMass*elementGravity;
	    centerOfCentrifugal +=
	    	e.elementMass*elementCentrifugal*elementHeight;
	    sumOfCentrifugal += e.elementMass*elementCentrifugal;
	}
        //elements[0].debug = true;   //TODO
	if (load>0.0) {
	    //If there is any load left over when we get to the end of the
	    //cable, we need to compensate for it with a counterweight.
	    double cwHeight = radius+length;  //cable top altitude
	    double cwGravity = gMass/(cwHeight*cwHeight); //c.top gravity
	    double cwCentrifugal = radial2*cwHeight;    //cable top centrifugal
	    double cwAcceleration = cwGravity - cwCentrifugal;
	    if (cwAcceleration==0.0) {
	        cwMass = Double.POSITIVE_INFINITY;
	    } else {
	        cwMass = -load/cwAcceleration;
		//Since we are out beyond synchronous orbit (we assume),
		//the net acceleration (cwAcceleration) is negative (away
		//from the planet), thus the negative sign to get positive mass.
		//If our cable is shorter than synchronous orbit, we would need
		//a negative mass to keep it from falling down, so our cwMass
		//will be negative.
	    }
	    centerOfMass += cwMass*cwHeight;
	    centerOfGravity += cwMass*cwGravity*cwHeight;
	    sumOfGravity += cwMass*cwGravity;
	    centerOfCentrifugal += cwMass*cwCentrifugal*cwHeight;
	    sumOfCentrifugal += cwMass*cwCentrifugal;
	} else {
	    cwMass = 0.0;	//no counterweight required
	}
	centerOfMass /= (mass+cwMass);
	centerOfMass -= radius;
	centerOfGravity /= sumOfGravity;
	centerOfGravity -= radius;
	centerOfCentrifugal /= sumOfCentrifugal;
	centerOfCentrifugal -= radius;

        calculateStressColorMap();

        //Calculate the initial accelerations for saving
        calculateStresses(0,0);
        calculateAccelerations(0,0,0.,0.);
	for (int i=0; i<numElements; i++) {
            elements[i].saveInitial();    //save the initial values
        }
    }

    private void calculateStressColorMap() {
        stressColorMap = new Color[STRESS_COLOR_COUNT];

        int cx = 0;
        stressColorMap[cx++] = Color.black;        //no stress (probably broken)

        int okStressCount = (int)(STRESS_COLOR_COUNT/safetyFactor);
        if (okStressCount>=STRESS_COLOR_COUNT)
            okStressCount = STRESS_COLOR_COUNT - 1;
        int cx0 = cx;
        float a = (float)okStressCount;
        for ( ; cx<okStressCount-1; cx++) {
            float z = ((float)(cx-cx0))/a;
            float zi = 1.0f - z;
            //Under the design load, color goes from blue to green
            stressColorMap[cx] = new Color(0.0f, z, zi);
        }
        stressColorMap[cx++] = Color.yellow;      //at the design load
        cx0 = cx;
        a = (float)(STRESS_COLOR_COUNT - cx0);
        for ( ; cx<STRESS_COLOR_COUNT; cx++) {
            float z = ((float)(cx-cx0))/a;
            float zi = 1.0f - z;
            //Above the design load, color goes from pink to red
            stressColorMap[cx] = new Color(0.5f+z/2.0f,0.0f,0.0f);
        }
    }

    /** Reset the simulation of our cable. */
    protected void resetSimulation() {
	recalculateElements();
    }

    /** Simulate the motion of our cable in the indicated amount of time.
     * @param simInterval The step size to simulate.
     * @param algorithm The algorithm to use, one of SpelCanvas.ALG_*.
     */
    protected void simulateOneStep(double simInterval, int algorithm,
            double dampingFactor, double cwDerating, double groundDrag) {
	if (elements==null)
	    return;
        switch (algorithm) {
        default:
        case SpelCanvas.ALG_FORWARD:
            //Using a simple Euler forward-integration does not work well,
            //the element positions are consistently miscalculated by a small
            //amount, which adds up and causes the cable to break.
            //But it is the simplest approach.
            calculateStresses(0,0);
                //calculate stresses given current position
            calculateAccelerations(0,0,cwDerating,groundDrag);
                //calculate accelerations based on the above stresses
            calculatePositions(simInterval,0,0,0,dampingFactor);
            break;
        case SpelCanvas.ALG_FORWARD_CIRCULAR:
        case SpelCanvas.ALG_TRAPEZOID:
            //With the trapezoidal approach, we first estimate the
            //accelerations at the mid-point of the interval (which will
            //typically be a bit higher than the real value), then apply
            //that set of values over the entire interval (rather than applying
            //the initial values over the entire interval).
            calculateStresses(0,0);
                //calculate stresses given current position
            calculateAccelerations(0,0,cwDerating,groundDrag);
                //calculate accelerations based on the above stresses
            double simFraction = (algorithm==SpelCanvas.ALG_FORWARD_CIRCULAR)?
                    FWD_CIRCLE_FRACTION:2.0;
            double simFracInterval = simInterval/simFraction;
            calculatePositions(simFracInterval,0,1,0,0.0);
                //calculate the position at the halfway point based on the
                //current location (0), into a temporary store (1)
                //for new location,
                //at the mid-point of this simluation step,
                //using the initial acceleration (0)
            calculateStresses(1,simFracInterval);
                //recalculate the stresses given the mid-point locations
            calculateAccelerations(1,simFracInterval,cwDerating,groundDrag);
                //recalculate acceleration based on the mid-point stresses
            int whichAccel = (algorithm==SpelCanvas.ALG_FORWARD_CIRCULAR)?-1:1;
                //the -1 tells calculatePositions to use the circular
                //acceleration algorithm.
            calculatePositions(simInterval,0,0,whichAccel,dampingFactor);
                //calculate the new positions based on the original positions
                //and veolcity, but using the new mid-point acceleration
            break;
        }

        //Debug
        //printMaxDrifts(4);
    }

    /** Calcuate the stress on each element.
     * The stress is based on the separation between the element and
     * the one below it as compared to the unloaded length of the element.
     * If the stress exceeds the strength of the material, the element
     * is marked as broken.
     */
    protected void calculateStresses(int whichDyn, double simOffset) {
    	for (int i=0; i<numElements; i++) {
            DVector below;
	    if (i==0) {
                below = planet.getBase(simOffset); //connected to base station
	    } else {
                below = elements[i-1].dyn[whichDyn].d;
                        //connected to element below
	    }
	    if (!elements[i].dyn[whichDyn].broken) {
	        elements[i].calculateStress(below,whichDyn);
		if (elements[i].dyn[whichDyn].broken) {
		    if (whichDyn==0) {
                        if (debugOutputEnabled || elements[i].debug) {
                            System.out.println("Element "+i+" just broke");
                        }
                        //numBrokenElements++;
                    }
		}
	    }
	}
    }

    /** Calculate the acceleration on each element.
     * The acceleration for each element is the vector sum of the contributions
     * from the connections to the elements above and below, and gravity.
     */
    protected void calculateAccelerations(int whichDyn, double simOffset,
            double cwDerating, double groundDrag) {
    	for (int i=0; i<numElements; i++) {
	    CableElement above = null;
	    if (i<numElements-1)
	        above = elements[i+1];
            DVector below;
	    if (i==0) {
	        below = planet.getBase(simOffset); //base station coordinates
	    } else {
	        below = elements[i-1].dyn[whichDyn].d;  //element below us
	    }
	    elements[i].calculateAcceleration(below,above,whichDyn,
                    cwDerating, groundDrag);
	}
    }

    /** Caculate the new position of each element.
     * The velocity is updated based on the acceleration, and
     * the position is updated based on the velocity.
     */
    protected void calculatePositions(double simInterval,
            int fromDyn, int toDyn, int whichAccel, double dampingFactor) {
    	for (int i=0; i<numElements; i++) {
            boolean wasGrounded = elements[i].dyn[toDyn].grounded;
	    elements[i].updatePosition(simInterval, fromDyn, toDyn, whichAccel,
                    dampingFactor);
            boolean isGrounded = elements[i].dyn[toDyn].grounded;
            if (!wasGrounded && isGrounded)
                numGroundedElements++;
	}
    }

    /** Draw the cable. */
    public void draw(Graphics2D g, double scale) {
    	if (elements==null)
	    return;		//nothing to draw
        int start = -1;         //below element 0 is the base station
        int e=0;
        for (; e<=numElements; e++) {
            if (e>=numElements || elements[e].dyn[0].broken) {
                drawSegment(g, scale, start, e-1);
                start = e;
            }
        }
    }

    //Draw a single fully connected segment of cable elements
    private void drawSegment(Graphics2D g, double scale, int start, int end) {
        if (end<0)
            return;     //nothing to draw
        if (start==end) {
            //special case for drawing a single point
            drawSection(g, scale, start, end);
            return;
        }
        double uLength = elements[end].unloadedLength;
            //get the original length of one element, assume same for all
        double pixelLength = scale*uLength;
            //the number of pixels long one element is
        double optimalPixelLength = 10.0;
            //the size in pixels of the smallest piece we want to draw
        int numDraw = (int)Math.round(optimalPixelLength/pixelLength);
            //the number of elements we should consider at one time.
        if (numDraw==0)
            numDraw++;
        int sectionCount = (end - start) / numDraw + 1;
        int elemsPerSection = (end - start) / sectionCount;
        if (elemsPerSection==0)
            elemsPerSection = 1;
        for (int s0=start; s0<end; s0+=elemsPerSection) {
            int e0 = s0 + elemsPerSection;
            if (e0>end)
                e0 = end;
            drawSection(g, scale, s0, e0);
        }
    }

    //Draw one section of connected cable, whose length we have already
    //determined is sufficiently small.
    private void drawSection(Graphics2D g, double scale, int start, int end) {
        DVector startPoint;
        DVector endPoint = elements[end].dyn[0].d;
        if (start<0)
            startPoint = planet.getBase();
        else
            startPoint = elements[start].dyn[0].d;
	int x0 = (int)(startPoint.x * scale);
	int y0 = (int)(startPoint.y * scale);
	int x1 = (int)(endPoint.x * scale);
	int y1 = (int)(endPoint.y * scale);
if (startPoint.getRadius()<planet.getRadius()*.99) {
  System.out.println("p0<radius");
}
if (endPoint.getRadius()<planet.getRadius()*.99) {
  System.out.println("p1<radius");
}
        //color for this piece is based on the average strain on the elements
        Color thisColor = getStressColor(start,end);
        Color oldColor = g.getColor();
        g.setColor(thisColor);
	g.drawLine(x0,y0,x1,y1);
        g.setColor(oldColor);
    }

    private Color getStressColor(int start, int end) {
        if (end<0)
            end = 0;
        if (start<0)
            start = 0;
        double maxStress = material.getYieldStrength();
        double totalStress = 0.0;
        for (int i=start; i<=end; i++) {
            totalStress += elements[i].dyn[0].stress;
        }
        double averageStress = totalStress/(end-start+1);
        double stressFraction = averageStress/maxStress;
        if (stressFraction<0)
        if (stressFraction>1.0)
            stressFraction = 1.0;       //the breaking point
        int stressIndex =
                (int)(stressFraction*(double)(stressColorMap.length-1));
        return stressColorMap[stressIndex];
    }

    /** Get the position vectors for all elements. */
    public DVector[] getPositions() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getPosition();
        }
        return dd;
    }

    /** Get the velocity vectors for all elements. */
    public DVector[] getVelocities() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getVelocity();
        }
        return dd;
    }

    /** Get the acceleration vectors for all elements. */
    public DVector[] getAccelerations() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getAcceleration();
        }
        return dd;
    }

    /** Get the ideal position vectors for all elements. */
    public DVector[] getIdealPositions() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getIdealPosition();
        }
        return dd;
    }

    /** Get the ideal velocity vectors for all elements. */
    public DVector[] getIdealVelocities() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getIdealVelocity();
        }
        return dd;
    }

    /** Get the ideal acceleration vectors for all elements. */
    public DVector[] getIdealAccelerations() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getIdealAcceleration();
        }
        return dd;
    }

    /** Get the position drift vectors for all elements. */
    public DVector[] getPositionDrifts() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getPositionDrift();
        }
        return dd;
    }

    /** Get the velocity drift vectors for all elements. */
    public DVector[] getVelocityDrifts() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getVelocityDrift();
        }
        return dd;
    }

    /** Get the acceleration drift vectors for all elements. */
    public DVector[] getAccelerationDrifts() {
        DVector[] dd = new DVector[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].getAccelerationDrift();
        }
        return dd;
    }

    /** Get the stresses for all elements. */
    public double[] getStresses() {
        double[] dd = new double[elements.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = elements[i].dyn[0].stress;
        }
        return dd;
    }

    /** Get the broken flag for all elements. */
    public boolean[] getBroken() {
        boolean[] bb = new boolean[elements.length];
        for (int i=0; i<bb.length; i++) {
            bb[i] = elements[i].dyn[0].broken;
        }
        return bb;
    }

    /** Get the elements with the largest amount of position drift.
     * @param count The number of elements to return.
     */
    public CableElement[] getMaxPositionDrifts(int count) {
        if (count<=0)
            return null;
        if (count>elements.length)
            count = elements.length;
        CableElement[] maxes = new CableElement[count];

        for (int e=0; e<elements.length; e++) {
            CableElement elem = elements[e];
            DVector drift = elem.getPositionDrift();
            double driftMag = drift.getRadius();

            for (int x=0; x<maxes.length; x++) {
                if (maxes[x]==null) {
                    maxes[x] = elem;
                    break;
                }
                if (driftMag > maxes[x].getPositionDrift().getRadius()) {
                    System.arraycopy(maxes,x,maxes,x+1,maxes.length - x - 1);
                    maxes[x] = elem;
                    break;
                }
                //not larger than this entry, continue on to the next one
            }
        }
        return maxes;
    }

    /** Print the maximum drift values. */
    public void printMaxPositionDrifts(int count) {
        CableElement[] elems = getMaxPositionDrifts(count);
        for (int e=0; e<count; e++) {
            elems[e].dumpPositionDrift();
        }
    }

    /** For debugging, dump the first and last cable elements. */
    public void dumpSelectedElements(double time, int algorithm) {
    	if (elements==null) {
	    System.out.println("No elements");
	    return;
	}
	System.out.println("t="+time);
        System.out.println("PlanetBase="+planet.getBase().toString());
	elements[0].dump(algorithm);
	elements[1].dump(algorithm);
	elements[numElements/2].dump(algorithm);
	elements[numElements-1].dump(algorithm);
    }
}

//end
