package net.jimmc.spelsim;

import java.text.MessageFormat;

/** A piece of the cable when modeling it as a set of discreet pieces. */
class CableElement {
    private static double maxA = 0;     //record maximum element acceleration
    private static double maxV = 0;     //record maximum element velocity
    boolean debug;      //true to debug on this element
    Cable cable;	//the cable of which this element is a part
    int index;		//0 for the element closest the planet, 1 for next
    double area;	//cross-sectional area of this element (unloaded)
    double elementMass; //mass of this element
    double unloadedLength;	//unloaded length of this element
    double Ld;	//unloadedLength times density (see calculateAcceleration)
    double r0;  //the original r value of this element
    CableDynamic[] dyn;      //The dynamic data for this element
    CableDynamic initial;       //The values of the cable dynamics at t=0

    class CableDynamic {
        double stress;	//stress on this element.
        DVector d;          //location of the center of this element
        DVector v;		//velocity of this element
        DVector a;		//acceleration for one simulation step
        boolean broken;	//true if we are not attached to the element below us
        boolean grounded;       //true if this segment has hit the ground
        double groundAngleOffset;
            //The position on the planet, relative to the planet's 0-angle,
            //when the cable landed.

        public CableDynamic() {
            d = new DVector(0.0,0.0);
            v = new DVector(0.0,0.0);
            a = new DVector(0.0,0.0);
        }

        public CableDynamic(CableDynamic source) {
            d = new DVector(source.d.x,source.d.y);
            v = new DVector(source.v.x,source.v.y);
            a = new DVector(source.a.x,source.a.y);
            stress = source.stress;
        }
    }

    CableElement(Cable c) {
    	cable = c;
        dyn = new CableDynamic[2];
        for (int i=0; i<dyn.length; i++)
            dyn[i] = new CableDynamic();
    }

    protected void saveInitial() {
        initial = new CableDynamic(dyn[0]);
    }

    /** Calculate the stress between this element and the one below.
     * If the stress exceeds the strength of the material,
     * set the broken flag.
     * @param otherPoint The location of the attachment below us.
     * @return True if this element just broke.
     */
    public void calculateStress(DVector otherPoint, int whichDyn) {
	//calculate stress based on distance from our x/y to the
	//specified x/y.
        if (debug) {
//System.out.println("debugFlag set for ["+index+"]");    //breakpoint hook
        }
        /*
        if (dyn[whichDyn].grounded) {
            //If grounded, no need to calculate this stuff, since pos is fixed.
            dyn[whichDyn].stress = 0.0;
            return;
        }
        */
        DVector delta = dyn[whichDyn].d.minus(otherPoint);
        double distance = delta.length();
	double strain = distance/unloadedLength - 1.0;
	if (strain<0) {
	    //no tension on this cable element, so no stress
	    dyn[whichDyn].stress = 0.0;
	    return;
	}
	Material material = cable.material;
	double youngs = material.getYoungsModulus();
	dyn[whichDyn].stress = (youngs==0.0)?0.0:(strain*youngs);
/*
System.out.println("["+index+"] "+
"strain="+Debug.numToString(strain)+
", stress="+Debug.numToString(dyn[whichDyn].stress));
*/
	if (dyn[whichDyn].stress > material.getYieldStrength()) {
                    //TODO - for now, use yield strength as breaking strength
	    //too much stress, this element just broke
            if (debug || cable.isDebugOutputEnabled()) {
                System.out.println("["+index+"]["+whichDyn+"] break: "+
                  "strain="+Debug.numToString(strain)+
                  ",youngs="+Debug.numToString(youngs)+
                  ",stress="+Debug.numToString(dyn[whichDyn].stress)+
                  ",strength="+Debug.numToString(material.getYieldStrength()));
            }
	    dyn[whichDyn].broken = true;
	    dyn[whichDyn].stress = 0.0;	//once broken, no more stress
	}
    }

    /** Calculate the acceleration on this element.
     * The acceleration on the element is computed as the sum of the forces
     * contributed by our connection to the below elements, the connection
     * of the above element to us, and the force due to gravity.
     * We assume for this calculation that the cross sectional area of the
     * cable elements above and below us are the same as ours, which assumption
     * is close to true if the taper is small and the number of elements is
     * large.
     * By making this assumption, we can do calculations based on acceleration
     * rather than force, and avoid the intermediate step of calculating
     * the actual forces on the element.
     * @param below The location of the element below us, or of the base
     *        station if we are the bottom element.
     * @param above The element above us, or null if we are the top element.
     * @param whichDyn Which set of dynamic data to modify.
     */
    public void calculateAcceleration(DVector below, CableElement above,
            int whichDyn, double cwDerating, double groundDrag) {
        if (debug) {
//System.out.println("debugFlag set for ["+index+"]");    //breakpoint hook
        }
        if (dyn[whichDyn].grounded) {
            dyn[whichDyn].a = new DVector(0.0,0.0);
                //no acceleration due to graavity if we are grounded
        } else {
            //Start by calculating the acceleration due to gravity
            double planetMass = cable.planet.getMass();
            double gMass = Physics.G*planetMass;
                            //TODO - put gMass into cable instance?
            double heightSquared = dyn[whichDyn].d.lengthSquared();
            double height = Math.sqrt(heightSquared);
            double ag = -gMass/heightSquared;
                    //acceleration is negative, i.e. towards (0,0)
            dyn[whichDyn].a = dyn[whichDyn].d.times(ag/height);
                //make acceleration point to origin, which is center of planet
        }
	//Add in the acceleration due to stress on our element
	if (!dyn[whichDyn].broken) {
	    double effectiveStress;
	    if (above==null) {
	        //This is the topmost segment, so we add in the effect of
		//the counterweight by reducing the acceleration due to the
		//stress of the below segment by the ratio of the element
		//mass to the element mass plus the effective CW mass.
                double cwMass = cable.cwMass*(1.0-cwDerating);
	        double massRatio = elementMass / (elementMass + cwMass);
		effectiveStress = dyn[whichDyn].stress*massRatio;
	    } else {
	        effectiveStress = dyn[whichDyn].stress;
            }
	    DVector ss = getAcceleration(dyn[whichDyn].d,below,effectiveStress);
            dyn[whichDyn].a = dyn[whichDyn].a.plus(ss);
	}
	if (above!=null && !above.dyn[whichDyn].broken) {
	    //The segment above us is a different size, so the same amount
	    //of stress on it will cause a slightly different acceleration
	    //on our segment.  We multiply the stress of that segment by
	    //the ratio of its size to ours in order to account for this.
	    double sizeRatio = above.area/this.area;
	    DVector ss = getAcceleration(dyn[whichDyn].d,above.dyn[whichDyn].d,
                    above.dyn[whichDyn].stress*sizeRatio);
            dyn[whichDyn].a = dyn[whichDyn].a.plus(ss);
	}
        if (dyn[whichDyn].grounded & groundDrag!=0.0) {
            //If we are grounded, add in the deceleration due to drag.
            DVector groundV = cable.planet.getGroundVelocity(dyn[whichDyn].d);
            DVector deltaV = groundV.minus(dyn[whichDyn].v);
                //(negative of) element velocity relative to the ground
            DVector dv = deltaV.times(groundDrag);
            dyn[whichDyn].a = dyn[whichDyn].a.plus(dv);
        }
        if (dyn[whichDyn].a.getRadius()>maxA) {
            maxA = dyn[whichDyn].a.getRadius();
            //System.out.println("New maxA="+maxA);
        }
    }

    /** Get the acceleration due to a neighboring element. */
    protected DVector getAcceleration(DVector thisD, DVector other,
            double stress) {
    	//The acceleration due to the stress on a neighboring segment is a=F/m.
	//where force F=stress*Area and mass m=Length*Area*Density, so
	// F/m = (stress*Area) / (Length*Area*Density) = stress/(Length*Density)
	//We use the unloaded length to calculate the element mass.
        DVector delta = other.minus(thisD);
        double dd = delta.length();
        DVector ss = delta.times(stress/dd/Ld);
        return ss;
    }

    /** Update our position.
     * Assumes acceleration has already been calculated.
     * Also updates our velocity.
     */
    public void updatePosition(double simInterval,
            int fromDyn, int toDyn, int whichAccel, double dampingFactor) {
        if (debug) {
//System.out.println("debugFlag set for ["+index+"]");    //breakpoint hook
        }
        CableDynamic fDyn = dyn[fromDyn];
        CableDynamic tDyn = dyn[toDyn];
        /*
        if (tDyn.grounded) {
            //If we are grounded, then our position is attached to the
            //planet as it rotates.
            tDyn.d = DVector.newPolar(tDyn.d.getRadius(),
                    cable.planet.getPrimeAngle()+tDyn.groundAngleOffset);
            return;
        }
        */

        if (whichAccel==-1) {
            //We have been told to use our forward circular algorithm.
            //We assume an acceleration with a fixed magnitude, that being
            //the magnitude of dyn[0].a.  We then look at the angular
            //difference between dyn[0].a and dyn[1].a, ignoring the
            //magnitude of dyn[1].a.  We assume dyn[1].a represents the
            //acceleration at a fractional amount of the interval, so we
            //multiply that number to get the assumed angle of change for
            //the acceleration vector during this interval.  We further
            //assume a constant angular velocity for the changing
            //acceleration vector.
            double theta = dyn[1].a.getAngle() - dyn[0].a.getAngle();
            while (theta>Math.PI)
                theta -= 2*Math.PI;
            while (theta<-Math.PI)
                theta += 2*Math.PI;
            theta *= Cable.FWD_CIRCLE_FRACTION;
                    //multiply the difference by the fraction
                    //of the step that we used to calcalulate the
                    //values in dyn[1] to get the amount of change
                    //over the full simulation interval.
            //Keep theta centered around zero; we assume the angular
            //change is less than PI, if it is greater we will be
            //completely wrong in our calculations.
            //Since we multiplied by FWD_CIRCLE_FRACTION, we could have
            //a very large number for theta at this point.
            theta = Math.IEEEremainder(theta,2*Math.PI);
            while (theta>Math.PI)
                theta -= 2*Math.PI;
            while (theta<-Math.PI)
                theta += 2*Math.PI;

            //If the angle is too small, assume the direction of
            //acceleration is constant (our calculations divide by theta,
            //so they will fail if theta is zero).
            if (Math.abs(theta)<1E-20) {
                //use constant acceleration algorithm
                updatePosition(simInterval, fromDyn, toDyn, 0, dampingFactor);
                return;
            }

            DVector baseA = dyn[0].a;
            DVector normalA = baseA.normal();
            //We are the acceleration is constant magnitude with a
            //constantly changing angle, so
            //  a sub x = k * cos(w * t)
            //  a sub y = k * sin(w * t)
            //Where w (omega) is an angular velocity such that
            //  theta = w * dt          (dt = delta t)
            //theta being the angle we calculated above, dt being the
            //simulation interval as passed in to us (simInterval).
            double w = theta/simInterval;       //omega

            //We use a rotated coordinate system with dyn[0].a defining
            //the x axis and its normal defining the y axis, and we
            //assume t=0 for the beginning of the interval.
            //Thus for our base acceleration vector dyn[0].a, we have
            //cos(w*t)==1 and sin(w*t)==0 (both at t==0).

            //The change in velocity is the first integral of the acceleration,
            //  delta v sub x = k/w * sin(w * t);
            //  delta v sub y = k/w * -cos(w * t);
            //where k is the magnitude of the acceleraiton vector dyn[0].a.

            //For velocity, evaluating the integral from 0 to delta t gives
            //  delta v sub x = k/w * (sin(w * dt) - 0)
            //                  = (k*dt)*(sin(theta)/theta)
            //  delta v sub y = k/w * (-cos(w * dt) - -1)
            //                  = (k*dt)*((1 - cos(theta))/theta);
            //The x term represents the contribution of the base acceleration
            //vector, the y term represents the contribution of the normal.
            //Since both vectors have the same magnitude (k), we multiply
            //the base vector by sin(theta)/theta, the normal by
            //(1-cos(theta))/theta, add them, and multiply by dt.
            DVector deltaV = (baseA.times(Math.sin(theta)/theta).plus(
                    normalA.times((1.0-Math.cos(theta))/theta))).times(
                    simInterval);

            //The change in position is the second integral of the
            //acceleration,
            //  delta x = k/(w*w) * -cos(w * t);
            //  delta y = k/(w*w) * -sin(w * t);

            //For position, evaluating the integral from 0 to delta t gives
            //  delta x = k/(w*w) * (-cos(w*dt) - -1)
            //                  = (k*dt^2/theta)*((1 - cos(theta))/theta);
            //  delta y = k/(w*w) * (-sin(w*dt) - 0) =
            //                  = (k*dt^2/theta)*(-sin(theta)/theta)
            //We add a theta (w*dt) term to delta y to get our initial
            //conditions right (v=0):
            //  delta y = k/(w*w) * (-sin(w*dt) - 0) =
            //                  = (k*dt^2/theta)*(theta - sin(theta)/theta)
            DVector deltaD = (baseA.times((1.0 - Math.cos(theta))/theta).plus(
                    normalA.times((theta-Math.sin(theta))/theta))).times(
                    simInterval*simInterval/theta);

            //Position change is due to this intervals acceleration plus
            //the initial velocity times delta t.
            tDyn.d = fDyn.d.plus(fDyn.v.times(simInterval));
                    // d += v * t
            tDyn.d = tDyn.d.plus(deltaD);
                    // add in the change due to acceleration in this interval
            tDyn.v = fDyn.v.plus(deltaV);
        } else {
            //Assume constant acceleration over simInterval.
            //In simple forward euler, we use the initial acceleration
            //(dyn[0].a) throughout the simulation interval.
            //When doing trapezoidal, we use the acceleration at
            //the half-way point (as calculated using simple forward euler)
            //for the entire interval.
            tDyn.d = fDyn.d.plus(fDyn.v.times(simInterval));
                    // d += v * t
            tDyn.d = tDyn.d.plus(
                    dyn[whichAccel].a.times(0.5*simInterval*simInterval));
                    //d += 1/2 * a * t^2
            tDyn.v = fDyn.v.plus(dyn[whichAccel].a.times(simInterval));
                    //new velocity
        }

        //Use damping factor to damp the velocity.
        //This is to counter the instability artifacts introduced by
        //the use of finite element simulation.
        if (dampingFactor!=0.0 && !tDyn.broken) {
            DVector below = null;
            DVector vBelow = null;
            if (index==0) {
                below = cable.planet.getBase(simInterval);
                vBelow = cable.planet.getBaseVelocity(simInterval);
            } else {
                below = cable.elements[index-1].dyn[toDyn].d;
                vBelow = cable.elements[index-1].dyn[toDyn].v;
            }
            DVector deltaD = tDyn.d.minus(below);
            DVector deltaV = tDyn.v.minus(vBelow);
            //Get the magnitude of the component of the velocity which
            //is in the direction of deltaD.
            double vBelowMag = deltaV.dot(deltaD.unit());
            double vDampMag = vBelowMag * dampingFactor;
            //Turn it into a vector pointing in the opposite direction as deltaD
            DVector vDamp = DVector.newPolar(-vDampMag,deltaD.getAngle());
            //Add the damping velocity back in
            tDyn.v = tDyn.v.plus(vDamp);
        }

        if (tDyn.v.getRadius()>maxV) {
            maxV = tDyn.v.getRadius();
            //System.out.println("New maxV="+maxV);
        }

        //After updating the position, see if we hit the planet.
        //We assume our velocity is small enough that we won't go
        //all the way through in a single simulation step.
        if (tDyn.d.getRadius()<=cable.planet.getRadius()) {
            //We use a pretty simple model when we hit the ground:
            //we just stick to the surface at about the point we hit.
            //We could look at the current velocity and calculate the
            //point along that trajectory that we hit, but we will
            //take the simpler approach of just changing the radius
            //to put the point on the surface, assuming the lateral
            //velocity would have caused it to slide to about that point.
            //In reality, our crash simulation is pretty weak - for
            //example, we have no modeling of the atmosphere, which
            //would be a significant effect, and we don't try to model
            //any friction on the surface, we just assume it is very sticky.
            tDyn.d = DVector.newPolar(
                    cable.planet.getRadius(),tDyn.d.getAngle());
            tDyn.groundAngleOffset = tDyn.d.getAngle() -
                    cable.planet.getPrimeAngle();
            if (tDyn.groundAngleOffset<0.0)
                tDyn.groundAngleOffset += 2.0*Math.PI;
            //tDyn.broken = true;
            tDyn.grounded = true;
        }
    }

    public int getIndex() {
        return index;
    }

    public boolean isBroken() {
        return dyn[0].broken;
    }

    public void setBroken() {
        dyn[0].broken = true;
        dyn[0].stress = 0.0;    //no stress once broken
    }

    public boolean isGrounded() {
        return dyn[0].grounded;
    }

    //Get the current position of this element
    public DVector getPosition() {
        return dyn[0].d;
    }

    //Get the current velocity of this element
    public DVector getVelocity() {
        return dyn[0].v;
    }

    //Get the current acceleration of this element
    public DVector getAcceleration() {
        return dyn[0].a;
    }

    //Get the ideal position of this element based on its initial position
    //and the rotation of the planet.
    public DVector getIdealPosition() {
        return initial.d.rotated(cable.getPlanet().getPrimeAngle());
    }

    //Get the ideal velocity of this element based on its initial velocity
    //and the rotation of the planet.
    public DVector getIdealVelocity() {
        return initial.v.rotated(cable.getPlanet().getPrimeAngle());
    }

    //Get the ideal acceleration of this element based on its
    //initial acceleration and the rotation of the planet.
    public DVector getIdealAcceleration() {
        return initial.a.rotated(cable.getPlanet().getPrimeAngle());
    }

    /** Get the position drift for this element, which is the difference
     * between the current position (dyn[0].d) and the ideal position.
     */
    public DVector getPositionDrift() {
        return getPosition().minus(getIdealPosition());
    }

    /** Get the velocity drift for this element, which is the difference
     * between the current velocity (dyn[0].v) and the ideal velocity.
     */
    public DVector getVelocityDrift() {
        return getVelocity().minus(getIdealVelocity());
    }

    /** Get the acceleration drift for this element, which is the difference
     * between the current acceleration (dyn[0].a) and the ideal position.
     */
    public DVector getAccelerationDrift() {
        return getAcceleration().minus(getIdealAcceleration());
    }

    /** Dump the position drift of this element, for debugging. */
    public void dumpPositionDrift() {
    	System.out.print("Element "+index+" drift: ");
        System.out.println(getPositionDrift().toString());
    }

    /** For debugging, dump one element. */
    public void dump(int algorithm) {
    	System.out.println("Element "+index);
        dumpDyn(dyn[0],0);
        if (algorithm==SpelCanvas.ALG_TRAPEZOID)
            dumpDyn(dyn[1],1);
    }
    private void dumpDyn(CableDynamic cDyn, int dynNum) {
	double strain = cDyn.stress/cable.material.getYoungsModulus();
    	System.out.println("  Dyn "+dynNum+" "+
		"strain="+Debug.numToString(strain)+
		", stress="+Debug.numToString(cDyn.stress)+
                (cDyn.broken?" broken":""));
	System.out.println("    d="+cDyn.d.toString()+
            " e="+polarOffset(cDyn.d.x,cDyn.d.y,
                    r0,cable.planet.getBaseTheta()));
	System.out.println("    v="+cDyn.v.toString());
	System.out.println("    a="+cDyn.a.toString());
    }

    private String polarOffset(double x, double y, double r0, double theta0) {
        double r1 = Math.sqrt(x*x+y*y);
        double theta1 = Math.atan2(y,x);
        return Debug.polarPointToString(r1-r0,r0*(theta1-theta0));
    }
}

//end
