package net.jimmc.spelsim;

/** Immutable 2D vector. */
public class DVector {
    double x;   //vector values in rectangular coordinates
    double y;

  //Constructors
    /** Create a vector with values (0,0). */
    public DVector() {}

    /** Create a vector by specifying rectangular coordinates. */
    public DVector(double x, double y) {
        this.x = x;
        this.y = y;
    }

    /** Create a vector by specifying polar coordinates. */
    public static DVector newPolar(double r, double theta) {
        double x = r*Math.cos(theta);
        double y = r*Math.sin(theta);
        return new DVector(x,y);
    }

    /** Return a new vector which is this vector times a constant. */
    public DVector times(double k) {
        return new DVector(k*x, k*y);
    }

    /** Return a new vector which is the sum of this vector plus another one. */
    public DVector plus(DVector that) {
        return new DVector(this.x + that.x, this.y + that.y);
    }

    /** Return a vector which is this vector minus another one. */
    public DVector minus(DVector that) {
        return new DVector(this.x - that.x, this.y - that.y);
    }

    /** Get a new vector which is a rotated version of this vector. */
    public DVector rotated(double theta) {
        if (x==0.0 && y==0.0)
            return new DVector(x,y);
        return newPolar(getRadius(),getAngle()+theta);
    }
    
    /** Get a vector normal to this vector and of the same magnitude.
     * The normal vector will have an angle of PI/2 greater than the
     * original vector (modulo 2 PI).
     */
    public DVector normal() {
        return new DVector(-y, x);
    }

    /** Get a vector of unit length pointing in the same direction
     * as this vector.
     * If this vector is (0,0), then (0,0) is returned.
     */
    public DVector unit() {
        if (x==0.0 && y==0.0)
            return new DVector(0.0,0.0);
        return newPolar(1.0,getAngle());
    }

    /** Get the length squared of this vector. */
    public double lengthSquared() {
        return x*x+y*y;
    }

    /** Get the length of this vector. */
    public double length() {
        return Math.sqrt(x*x+y*y);
    }

    /** Get the x coordinate of this point. */
    public double getX() {
        return x;
    }

    /** Get the y coordinate of this point. */
    public double getY() {
        return y;
    }

    /** Get the radius of the current point in polar coordinates. */
    public double getRadius() {
        return Math.sqrt(x*x+y*y);
    }

    /** Get the angle of the current point in polar coordinates.
     * @return The angle in radians, with 0 being the x axis.
     *  If the point is (0,0), the returned value is 0.
     */
    public double getAngle() {
        if (x==0.0 && y==0.0)
            return 0.0;
        return Math.atan2(y,x);
    }

    /** Get the dot product of this vector and another vector. */
    public double dot(DVector that) {
        double angle = getAngle() - that.getAngle();
        return getRadius() * that.getRadius() * Math.cos(angle);
    }

  //Misc methods
    public String toString() {
        return Debug.pointToString(x,y);
    }
}
