/* InspectPanel.java
 *
 * Jim McBeath, November 11, 2007
 */

package net.jimmc.spelsim;

import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.TextArea;
import java.awt.TextField;
import java.util.EventObject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

/** The panel implementing the tab for the Inspect button. */
class InspectPanel extends Panel implements TabSelectListener {
    private FilteredElements filteredElements;

    public InspectPanel(Cable cable, SpelCanvas canvas) {
        super();

        filteredElements = new FilteredElements(cable);

        InspectFilterPanel filterPanel =
                new InspectFilterPanel(filteredElements);
        InspectElementPanel elementPanel =
                new InspectElementPanel(filteredElements);

        GridBagLayout gbl = new GridBagLayout();
        this.setLayout(gbl);
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.anchor = GridBagConstraints.NORTH;

        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.gridwidth = 1;
        gbc.weightx = 1;
        gbc.weighty = 1;
        this.add(filterPanel, gbc);

        gbc.gridx = 1;
        this.add(elementPanel, gbc);
    }

    public void tabSelected() {
        filteredElements.updateFilteredElements();
    }

    public void tabDeselected(){}       //do nothing
}

/** A class to manage the set of cable elements selected by the filtering
 * parameters.
 */
class FilteredElements {
    private Cable cable;
    private Boolean filterBroken;
    private Boolean filterGrounded;
    private int orderBy;
        final static int ORDER_BY_NUMBER = 0;    //default
        final static int ORDER_BY_DRIFT = 1;
        final static int ORDER_BY_STRESS = 2;
    private List listenerList;
    private CableElement[] acceptedElements = new CableElement[0];

    public FilteredElements(Cable cable) {
        this.cable = cable;
        listenerList = new ArrayList();
        updateFilteredElements();
    }

    /** Convert a string to a boolean (actually trinary) choice.
     * @param s One of the strings "true", "false", or blank.
     * @return Boolean.TRUE or FALSE, or null if the arg was blank.
     */
    public static Boolean getBooleanChoice(String s) {
        if (s==null || s.trim().equals(""))
            return null;
        return (new Boolean(s.equalsIgnoreCase("true")));
    }

    public void setFilterBroken(Boolean b) {
        if (sameBoolean(b,filterBroken))
            return;     //no change
        filterBroken = b;
        updateFilteredElements();
    }

    public void setFilterGrounded(Boolean b) {
        if (sameBoolean(b,filterGrounded))
            return;     //no change
        filterGrounded = b;
        updateFilteredElements();
    }

    public void setOrderBy(int orderBy) {
        this.orderBy = orderBy;
        updateFilteredElements();
    }

    protected void updateFilteredElements() {
        List acceptedList = new ArrayList();
        CableElement[] elements = cable.getElements();
        for (int e=0; e<elements.length; e++) {
            CableElement elem = elements[e];
            if (acceptElement(elem)) {
                acceptedList.add(elem);
            }
        }
        acceptedElements =
                (CableElement[])acceptedList.toArray(new CableElement[0]);
        Arrays.sort(acceptedElements,new ElementSorter(orderBy));
        FilteredElementsChangeEvent ev = new FilteredElementsChangeEvent(this);
        fireChangeEvent(ev);
    }

    /** True if we should include this element in our filtered set. */
    protected boolean acceptElement(CableElement elem) {
        if (filterBroken!=null) {
            if (filterBroken.booleanValue()!=elem.isBroken())
                return false;
        }
        if (filterGrounded!=null) {
            if (filterGrounded.booleanValue()!=elem.isGrounded())
                return false;
        }
        return true;
    }

    class ElementSorter implements Comparator {
        private int orderBy;
        public ElementSorter(int orderBy) {
            this.orderBy = orderBy;
        }
        public int compare(Object a, Object b) {
            CableElement elemA = (CableElement)a;
            CableElement elemB = (CableElement)b;
            switch (orderBy) {
            default:
            case ORDER_BY_NUMBER:
                break;          //number sort is after the switch statement
            case ORDER_BY_DRIFT:
                //Highest drift magnitude first
                double driftA = elemA.getPositionDrift().getRadius();
                double driftB = elemB.getPositionDrift().getRadius();
                if (driftB>driftA)
                    return 1;
                if (driftB<driftA)
                    return -1;
                break;
            case ORDER_BY_STRESS:
                //Highest stress first
                double stressA = elemA.dyn[0].stress;
                double stressB = elemB.dyn[0].stress;
                if (stressB>stressA)
                    return 1;
                if (stressB<stressA)
                    return -1;
                break;
            }
            //Lowest element number first
            return (elemA.getIndex() - elemB.getIndex());
        }
    }

    public CableElement[] getAcceptedElements() {
        return acceptedElements;
    }

    private boolean sameBoolean(Boolean b1, Boolean b2) {
        if (b1==b2)
            return true;        //include case where both are null
        if (b1==null || b2==null)
            return false;       //only one is null
        return b1.equals(b2);
    }

    public void addChangeListener(FilteredElementsChangeListener listener) {
        if (!listenerList.contains(listener))   //no dups
            listenerList.add(listener);
    }
    public void removeChangeListener(FilteredElementsChangeListener listener) {
        listenerList.remove(listener);
    }
    public void fireChangeEvent(FilteredElementsChangeEvent ev) {
        Iterator it = listenerList.iterator();
        while (it.hasNext()) {
            FilteredElementsChangeListener listener =
                    (FilteredElementsChangeListener)it.next();
            listener.filteredElementsChanged(ev);
        }
    }
}

/** A listener that gets called when the set of filtered elements change. */
interface FilteredElementsChangeListener {
    public void filteredElementsChanged(FilteredElementsChangeEvent ce);
}
class FilteredElementsChangeEvent extends EventObject {
    private FilteredElements filteredElements;
    public FilteredElementsChangeEvent(FilteredElements source) {
        super(source);
        this.filteredElements = source;
    }
    public FilteredElements getFilteredElements() {
        return filteredElements;
    }
}
/** A simple adapter for getting change notices from FilteredElements. */
abstract class FilteredElementsChangeAdapter
        implements FilteredElementsChangeListener {
    public FilteredElementsChangeAdapter(){}
    public abstract void filteredElementsChanged(FilteredElementsChangeEvent e);
}

/** The panel that has all of the selection parameters for
 * filtering cable elements.
 */
class InspectFilterPanel extends PartPanel {
    ChoiceField brokenChoice;
    ChoiceField groundedChoice;
    ChoiceField orderByChoice;
    FilteredElements filteredElements;
    private final String[] booleanChoices = { "     ", "true", "false" };

    public InspectFilterPanel(FilteredElements fe) {
        super(null);
        this.filteredElements = fe;

        addLabel("Filter Parameters");

        brokenChoice = addChoice("Broken",booleanChoices,null);
        brokenChoice.addChoiceListener(new ChoiceListener() {
            public void choiceSelected(String s) {
                Boolean b = FilteredElements.getBooleanChoice(s);
                filteredElements.setFilterBroken(b);
            }
        });

        groundedChoice = addChoice("Grounded",booleanChoices,null);
        groundedChoice.addChoiceListener(new ChoiceListener() {
            public void choiceSelected(String s) {
                Boolean b = FilteredElements.getBooleanChoice(s);
                filteredElements.setFilterGrounded(b);
            }
        });

        String[] orderChoices = { "Number", "Drift", "Stress" };
                //must match ORDER_BY_*
        orderByChoice = addChoice("Order By",orderChoices,null);
        orderByChoice.addChoiceListener(new ChoiceListener() {
            public void choiceSelected(String s) {
                filteredElements.setOrderBy(orderByChoice.getSelectedIndex());
            }
        });
    }
}

/** The panel that shows the specific element being inspected. */
class InspectElementPanel extends PartPanel {
    IntegerSpinnerField indexNumberField;
    Label indexRightLabel;
    IntegerField elementNumberField;
    ChoiceField brokenChoice;
    ChoiceField groundedChoice;
    TextField positionField;
    TextField velocityField;
    TextField accelerationField;
    TextField positionIdealField;
    TextField velocityIdealField;
    TextField accelerationIdealField;
    TextField positionDriftField;
    TextField velocityDriftField;
    TextField accelerationDriftField;
    TextField stressField;
    TextField massField;
    TextField areaField;
    FilteredElements filteredElements;
    CableElement[] elements;

    public InspectElementPanel(FilteredElements fe) {
    	super(null);
        this.filteredElements = fe;
        fe.addChangeListener(new FilteredElementsChangeAdapter() {
            public void filteredElementsChanged(FilteredElementsChangeEvent ev){
                FilteredElements fe = ev.getFilteredElements();
                setAcceptedElements(fe.getAcceptedElements());
//System.out.println("Element panel got FE change event");        //TODO debug
            }
        });

        addLabel("Filtered Elements");

	indexNumberField = addIntegerSpinnerField("Index","");
	indexNumberField.addEnterListener(new EnterListener() {
	    public void valueEntered(int n) {
                displayElementByIndex(n);
	    }
	});
	indexNumberField.setValue(0);
        indexRightLabel = getLastRightLabel();

	elementNumberField = addIntegerField("Number","");
	elementNumberField.setValue(0);
        elementNumberField.setEditable(false);

        int vFieldSize = 60;        //big enough to show our vector info
        int fFieldSize = 20;    //big enough for one double
        positionField = addTextField("Position",vFieldSize,"m");
        positionField.setEditable(false);
        velocityField = addTextField("Velocity",vFieldSize,"m/s");
        velocityField.setEditable(false);
        accelerationField = addTextField("Acceleration",vFieldSize,"m/(s^2)");
        accelerationField.setEditable(false);
        positionIdealField = addTextField("P Ideal",vFieldSize,"m");
        positionIdealField.setEditable(false);
        velocityIdealField = addTextField("V Ideal",vFieldSize,"m/s");
        velocityIdealField.setEditable(false);
        accelerationIdealField = addTextField("A Ideal",vFieldSize,"m/(s^2)");
        accelerationIdealField.setEditable(false);
        positionDriftField = addTextField("P Drift",vFieldSize,"m");
        positionDriftField.setEditable(false);
        velocityDriftField = addTextField("V Drift",vFieldSize,"m/2");
        velocityDriftField.setEditable(false);
        accelerationDriftField = addTextField("A Drift",vFieldSize,"m/(s^2)");
        accelerationDriftField.setEditable(false);
        stressField = addTextField("Stress",fFieldSize,"N/(m^2)");
        stressField.setEditable(false);
        massField = addTextField("Mass",fFieldSize,"kg");
        massField.setEditable(false);
        areaField = addTextField("Cross Section",fFieldSize,"m^2");
        areaField.setEditable(false);

        final String[] brokenChoices = { "true", "false" };
        brokenChoice = addChoice("Broken",brokenChoices,null);
        brokenChoice.setEnabled(false);

        final String[] groundedChoices = { "true", "false" };
        groundedChoice = addChoice("Grounded",groundedChoices,null);
        groundedChoice.setEnabled(false);

        setAcceptedElements(fe.getAcceptedElements());
        displayElementByIndex(0);
    }

    public void setAcceptedElements(CableElement[] elements) {
        this.elements = elements;
        String countStr = "/"+elements.length;
        indexRightLabel.setText(countStr);
        displayElementByIndex();      //redisplay the currently selected index
    }

    /** Read the Index field and display that element. */
    public void displayElementByIndex() {
        int n = indexNumberField.getIntValue();
        displayElementByIndex(n);
    }

    /** Display the element at the specified index. */
    public void displayElementByIndex(int n) {
        if (elements.length==0) {
            //nothing to display
            elementNumberField.setText("");
            positionField.setText("");
            velocityField.setText("");
            accelerationField.setText("");
            positionIdealField.setText("");
            velocityIdealField.setText("");
            accelerationIdealField.setText("");
            positionDriftField.setText("");
            velocityDriftField.setText("");
            accelerationDriftField.setText("");
            stressField.setText("");
            massField.setText("");
            areaField.setText("");
            brokenChoice.select(0);
            groundedChoice.select(0);
            return;
        }
        if (n<0)
            n = 0;
        if (n>=elements.length)
            n = elements.length-1;
        CableElement elem = elements[n];
        elementNumberField.setText(Integer.toString(elem.getIndex()));
        positionField.setText(elem.dyn[0].d.toString());
        velocityField.setText(elem.dyn[0].v.toString());
        accelerationField.setText(elem.dyn[0].a.toString());
        positionIdealField.setText(elem.getIdealPosition().toString());
        velocityIdealField.setText(elem.getIdealVelocity().toString());
        accelerationIdealField.setText(elem.getIdealAcceleration().toString());
        positionDriftField.setText(elem.getPositionDrift().toString());
        velocityDriftField.setText(elem.getVelocityDrift().toString());
        accelerationDriftField.setText(elem.getAccelerationDrift().toString());
        stressField.setText(Debug.numToString(elem.dyn[0].stress));
        massField.setText(Debug.numToString(elem.elementMass));
        areaField.setText(Debug.numToString(elem.area));
        brokenChoice.select(elem.isBroken()?"true":"false");
        groundedChoice.select(elem.isGrounded()?"true":"false");
    }
}

//end
