/* LineCanvas.java
 *
 * Jim McBeath, November 16, 2007
 */

package net.jimmc.spelsim;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Panel;
import java.awt.Point;
import java.awt.Rectangle;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

/** A simple drawing canvas to draw a set of line graphs. */
public class LineCanvas extends Panel {
    private static final int perLabelWidth = 130;
    private static final int perLabelHeight = 12;
    private static final int labelBaseY = 14;
    private static final int labelsPerLine = 1;

    private List dataSets = new ArrayList();
    private int[] drawingOrder = new int[0];
        //the last item in the list is the topmost
    private String magFmt = "{0,number,0.####E0}";

    public LineCanvas() {
        super();
        setBackground(new Color(200,200,200));
        addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                raiseDataSet(e.getPoint());
            }
        });
    }

    /** Get the current drawing order. */
    public int[] getDrawingOrder() {
        int[] d = new int[drawingOrder.length];
        System.arraycopy(drawingOrder,0,d,0,drawingOrder.length);
        return d;
    }

    /** Set the current drawing order.
     * The data must be
     */
    public void setDrawingOrder(int[] d) {
        //If the data is valid, we will be making our own copy.
        //Do it now to ensure the caller can't change it while we are
        //looking at it.
        int[] dd = new int[d.length];
        System.arraycopy(d,0,dd,0,d.length);
        if (dd.length!=drawingOrder.length)
            return;             //ignore it
        for (int i=0; i<dd.length; i++) {
            if (dd[i]<0 || dd[i]>=dd.length)
                return;         //a bad index, ignore everything
        }
        //TODO - we could check to ensure that we have exactly one of
        //each number in the array; but at least we have validated that
        //the data won't make us crash.
        drawingOrder = dd;
    }

    /** Raise to the top of the drawing order the data set whose label
     * is under the given point.
     */
    public void raiseDataSet(Point p) {
        int lx = p.x/perLabelWidth;
        int y = p.y - (labelBaseY - perLabelHeight);
        if (y<0)
            return;     //above the top label
        int ly = y/perLabelHeight;
        if (lx<0 || ly<0)
            return;     //no label there
        if (lx>=labelsPerLine)
            return;     //to far to the right, no label there
        int n = ly*labelsPerLine+lx;
        if (n>getDataSetCount())
            return;    //nothing there
        raiseDataSet(n);
    }

    /** Raise to the top of the drawing order the data set of the given index.
     */
    public void raiseDataSet(int n) {
        if (n<0 || n>=getDataSetCount())
            return;
        //Start by finding the location of this index in the current order
        for (int i=0; i<drawingOrder.length; i++) {
            if (drawingOrder[i]==n) {
                //Found it, move it to the end of the list
                if (i==drawingOrder.length-1)
                    return;     //already at the end, leave it alone
                int[] newDrawingOrder = new int[drawingOrder.length];
                System.arraycopy(drawingOrder,0,newDrawingOrder,0,i);
                System.arraycopy(drawingOrder,i+1,newDrawingOrder,i,
                        drawingOrder.length-i-1);
                newDrawingOrder[drawingOrder.length-1] = n;
                drawingOrder = newDrawingOrder;
                repaint(10);
                return;
            }
        }
        //not found, ignore it
    }

    public Dimension getPreferredSize() {
        return new Dimension(200,100);
    }

    /** Clear out all of the data sets.
     */
    public void clearDataSets() {
        dataSets = new ArrayList();
        drawingOrder = new int[0];
        repaint(1);
    }

    /** Add a data set.
     * @return The data set index.
     */
    public int addDataSet(String name, double[] data) {
        DataSet ds = new DataSet();
        ds.setName(name);
        ds.setData(data);
        dataSets.add(ds);
        int n = dataSets.size() - 1;
        int[] newDrawingOrder = new int[n+1];
        System.arraycopy(drawingOrder,0,newDrawingOrder,0,n);
        newDrawingOrder[n] = n;
        drawingOrder = newDrawingOrder;
        repaint(1);
        return n;
    }

    /** Add a data set of boolean values.
     * @return The data set index.
     */
    public int addDataSet(String name, boolean[] data) {
        double[] dd = new double[data.length];
        for (int i=0; i<dd.length; i++) {
            dd[i] = data[i]?1.0:0.0;    //convert boolean to double
        }
        return addDataSet(name,dd);
    }

    /** Add two data sets, one for the magnitude and one for the angle.
     * @return The data set index of the first data set, which is the
     *         magnitude.  The index of the angle dataset is this number
     *         plus one.
     */
    public int addPolarDataSets(String name, DVector[] data) {
        double[] mag = new double[data.length];
        double[] ang = new double[data.length];
        for (int i=0; i<data.length; i++) {
            mag[i] = data[i].getRadius();
            double a = data[i].getAngle();
            a = Math.IEEEremainder(a,2*Math.PI);
                //ensure the value is between 0 and 2 pi
            if (a>Math.PI)      //we want it between +- PI
                a -= 2*Math.PI;
            ang[i] = a;
        }
        int mn = addDataSet(name+"Mag",mag);
        int an = addDataSet(name+"Ang",ang);
        DataSet ads = getDataSet(an);
        ads.setMinimumRange(Math.PI);   //keep scale on angles constant
        return mn;
    }

    /** Add two data sets, one for x and one for y.
     * @return The data set index of the first data set, which is X.
     *         The index of the Y dataset is this number plus one.
     */
    public int addRectangularDataSets(String name, DVector[] data) {
        double[] x = new double[data.length];
        double[] y = new double[data.length];
        for (int i=0; i<data.length; i++) {
            x[i] = data[i].getX();
            y[i] = data[i].getY();
        }
        int xn = addDataSet(name+"X",x);
        int yn = addDataSet(name+"Y",y);
        return xn;
    }

    public synchronized int getDataSetCount() {
        return dataSets.size();
    }

    public synchronized DataSet getDataSet(int n) {
        return (DataSet)dataSets.get(n);
    }

    public void paint(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        Color oldColor = g.getColor();
        int dsCount = getDataSetCount();
        //Draw all the labels
        for (int n=0; n<dsCount; n++) {
            g.setColor(getColor(n));
            DataSet ds = getDataSet(n);
            double maxMag = ds.getMaximumMagnitude();
            drawLabel(g,n,ds.getName(),maxMag);
        }
        //Flip the coordinate system to be math-style, with (0,0) lower left
        g2.scale(1.0,-1.0);
        g2.translate(0,-getHeight());
        int planetSize = 10;    //diameter of "planet"
        //Get the drawing area into which we put the lines
        int xLeft = planetSize+2;         //distance from left side of window
        int xRight = 2;         //distance from right side of window
        int yBottom = 2;        //distance from bottom of window
        int yTop = 4+dsCount*perLabelHeight;
                //distance from top of window (space for labels)
        int drawAreaHeight = getHeight() - (yTop+yBottom);
        int drawAreaWidth = getWidth() - (xLeft+xRight);
        int yZeroLine = yBottom+drawAreaHeight/2;
        Rectangle drawArea = new Rectangle(xLeft,yBottom,
                drawAreaWidth,drawAreaHeight);
        g.setColor(Color.white);        //make the actual drawing area white
        g2.fill(drawArea);
        g.setColor(Color.black);
        //Draw the "planet"
        g.drawArc(xLeft-planetSize,yZeroLine-planetSize/2,
                planetSize,planetSize,0,360);
        //Draw the zero axis
        g.drawLine(xLeft,yZeroLine,getWidth()-xRight,yZeroLine);
        //Draw all the lines
        for (int i=0; i<getDataSetCount(); i++) {
            int n = drawingOrder[i];
            g.setColor(getColor(n));
            DataSet ds = getDataSet(n);
            double maxMag = ds.getMaximumMagnitude();
            drawDataLine(g,n,ds.getScalarData(),maxMag,drawArea);
        }
        g.setColor(oldColor);
    }

    /** Get the color to use for a data set, by index. */
    private Color getColor(int n) {
        if (n<0)
            return Color.black;
        if (n>=dataColors.length)
            return Color.black;
        return dataColors[n];
    }
    private Color[] dataColors = { Color.red, Color.green, Color.blue,
        Color.yellow };

    private void drawLabel(Graphics g, int n, String name, double maxMag) {
        int nn = n/labelsPerLine;
        int nr = n%labelsPerLine;
        int x = nr * perLabelWidth;
        int y = nn * perLabelHeight + labelBaseY;
        String s = getNameScaleString(name,maxMag);
        g.drawString(s,x,y);
    }

    /** Get a string with the name and the scale magnitude */
    private String getNameScaleString(String name, double mag) {
        Object[] dd = { new Double(mag) };
        return name+":"+MessageFormat.format(magFmt,dd);
    }

    private void drawDataLine(Graphics g, int n, double[] data, double maxMag,
            Rectangle drawArea) {
        int xZeroPixel = drawArea.x;
        int yZeroPixel = drawArea.y + drawArea.height/2;
        int xPixelScale = drawArea.width;
        int yPixelScale = drawArea.height/2;
        int xData[] = new int[data.length];
        int yData[] = new int[data.length];
        for (int i=0; i<data.length; i++) {
            xData[i] = xZeroPixel+(((i+1)*xPixelScale)/data.length);
            yData[i] = yZeroPixel+(int)((data[i]*yPixelScale)/maxMag);
        }
        g.drawPolyline(xData,yData,data.length);
    }
}
