/* Material.java
 *
 * Jim McBeath, March 14, 2003
 */

package net.jimmc.spelsim;

import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;

/** A description of a material.
 * We use a fairly simple two-modulus model of the material.
 * The first modulus is Young's Modulus, which is the stress/strain
 * for stresses less than the yield strength of the material.
 * In this region the deformation is assumed to be completely
 * elastic, so that the material returns to its former length
 * when the stress is removed.
 * Above the yield point, the deformation is inelastic, so that
 * the material stays stretched even when the stress is reduced.
 * The stress/strain ratio in this second region is typically much
 * less, so that the material lengthens more for a given addition amount
 * of stress as compared to in the elastic region.
 * Once the stress reaches the breakingStress, the material breaks.
 * At this point, it will have stretched by an amount specified by
 * the breaking strain.
 * The strain at the yield point can be calculated from the
 * yield strength and the Young's modulus.
 * The difference between the breaking strain and the yield point strain,
 * together with the difference between the breaking strength and the
 * yield strength, can be used to determine the stress/strain ratio
 * of the inelastic portion of the model.
 */
public class Material {
    /** Density of the material in kg/(m^3) */
    double density;

    /** Breaking tensile strength of the material in Pascals (N/(m^2)) */
    double breakingStrength;

    /** Yield tensile strength in Pascals */
    double yieldStrength;

    /** Young's modulus */
    double youngsModulus;

    /** Strain at breaking point. */
    double breakingStrain;

    /** A set of known material values. */
    static Map knownMaterials;

    /** Get a material with preset values.
     */
    public static Material getMaterial(String name) {
        if (knownMaterials==null) {
	    initKnownMaterials();
	}
	return (Material)knownMaterials.get(name);	//null if not found
    }

    /** Get the list of known material names. */
    public static String[] getMaterialNames() {
        if (knownMaterials==null) {
	    initKnownMaterials();
	}
        String[] names =
		(String[])knownMaterials.keySet().toArray(new String[0]);
	Arrays.sort(names);
	return names;
    }

    private static void initKnownMaterials() {
	knownMaterials = new HashMap();
	//          name
	//          strength  yield   young's  density  breaking-strain
	addMaterial("Aluminum",
		    110e6,    95e6,   70e9,    2710e0,  0.0);
        addMaterial("Carbon Nanotubes",
		    130e9,    0.0,    1.25e12, 1300e0,  0.0);
	addMaterial("Cotton",
		    225e6,    0.0,    7.9e9,   1540e0,  0.0);
        addMaterial("Kevlar",
		    3.6e9,    3.6e9,  90e9,    1440e0,  0.0);
	addMaterial("Nylon Fibre",
		    600e6,    0.0,    3.9e9,   1140e0,  0.0);
	addMaterial("Spider Silk",
		    1.1e9,    0.15e9, 7.9e9,   1300e0,  0);
        addMaterial("Steel",
		    1.3e9,    250e6,  200e9,   7850e0,  0);
    }

    //Young's modulus is stress/strain where stress is force/area (in units
    //of Pascals) and strain is the change of length of the material
    //(i.e. deltaLength/length, unitless).
    //In other words, the amount of change in the length of a material is
    //   deltaLength = length*stress/youngsModulus
    //as long as the stress is less than the yieldStrength of the material.

    //Hook's law: strain (change in length) is proportional to stress.
    //Also referred to as the ideal spring law, F = -k*s where s is the
    //change in length and k is the spring constant.

    //Poisson ratio (nu): the amount of shrinkage of a material in the lateral
    //dimension when a stress is applied.  Must be -1 < nu <= 0.5.
    //The Poisson ratio for most metals is between 0.25 and 0.35; steel is 0.33.
    //Rubber is almost 0.5; if exactly 0.5, it would be truly incompressible.
    //Cork is close to zero.

    /** Add a material to our list of known materials.
     * @param name The name of the material.
     * @param breakingstrength The breaking strength of the material,
     *                 in Pascals.
     *		       Applying more than this amount of stress will break it.
     * @param yieldStrength The yield strength of the material, in Pascals.
     *                 As long as the applied stress is lower than this
     *                 much, the material elongates according to the
     *                 Young's modulus and the change is elastic.
     *              .  Above this amount of stress, the
     *                 material elongates more than as given by the Young's mod.
     *                 If 0, the yield strength is the same as the
     *                 breaking strength.
     * @param youngsModulus The Young's Modulus of the material, in Pascals.
     * @param density The density of the material, in kg/(m^3).
     * @param breakingStrain The strain at the breaking point, unitless
     */
    private static void addMaterial(String name, double breakingStrength,
    		double yieldStrength, double youngsModulus, double density,
		double breakingStrain) {
        Material m = new Material();
	if (yieldStrength==0.0)
	    yieldStrength = breakingStrength;
	m.setBreakingStrength(breakingStrength);
	m.setYieldStrength(yieldStrength);
	m.setYoungsModulus(youngsModulus);
	m.setDensity(density);
	knownMaterials.put(name,m);
    }

    /** Create a material. */
    public Material() {}

    /** Set the density of the material.
     * @param density The density of the material in kg/m^3.
     */
    public void setDensity(double density) {
    	this.density = density;
    }

    /** Get the density of the cable.
     * @return The density of the cable in kg/m^3.
     */
    public double getDensity() {
    	return density;
    }

    /** Set the breaking strength of the material.
     * @param breakingStrength The breaking strength of the material in Pascals.
     */
    public void setBreakingStrength(double breakingStrength) {
    	this.breakingStrength = breakingStrength;
    }

    /** Get the breaking strength of the cable.
     * @return The breaking strength of the cable in Pascals.
     */
    public double getBreakingStrength() {
    	return breakingStrength;
    }

    /** Set the yield strength of the material.
     * @param yieldStrength The yield strength of the material in Pascals.
     */
    public void setYieldStrength(double yieldStrength) {
    	this.yieldStrength = yieldStrength;
    }

    /** Get the yield strength of the cable.
     * @return The yield strength of the cable in Pascals.
     */
    public double getYieldStrength() {
    	return yieldStrength;
    }

    /** Set the Young's Modulus of the material.
     * @param youngsModulus The Young's Modulus of the material in Pascals.
     */
    public void setYoungsModulus(double youngsModulus) {
    	this.youngsModulus = youngsModulus;
    }

    /** Get the Young's Modulus of the cable.
     * @return The Young's Modulus of the cable in Pascals.
     */
    public double getYoungsModulus() {
    	return youngsModulus;
    }

    /** Get the strain at the yield point. */
    public double getYieldStrain() {
        if (youngsModulus==0.0)
	    return 0.0;
    	return yieldStrength/youngsModulus;
    }
}

//end
