/* Planet.java
 *
 * Jim McBeath, March 12, 2003
 */

package net.jimmc.spelsim;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/** A representation of a planet. */
public class Planet {
    /** The mass in kg. */
    double mass;

    /** The radius in meters. */
    double radius;

    /** The angle of our prime meridian away from its starting angle,
     * in radians.
     */
    double primeAngle;

    /** The gravity at the surface in meters per second-squared.*/
    //double gravity;
    	//calculate from mass and radius

    /** The period of rotation in seconds. */
    double period;

    private Color planetColor = new Color(0.0f, 0.4f, 0.8f);

    /** A set of known planet values. */
    static Map knownPlanets;

    /** Get a planet with preset values.
     */
    public static Planet getPlanet(String name) {
        if (knownPlanets==null) {
	    initKnownPlanets();
	}
	return (Planet)knownPlanets.get(name);	//null if not found
    }

    /** Get the list of known planet names. */
    public static String[] getPlanetNames() {
        if (knownPlanets==null) {
	    initKnownPlanets();
	}
        String[] names = (String[])knownPlanets.keySet().toArray(new String[0]);
	Arrays.sort(names);
	return names;
    }

    private static void initKnownPlanets() {
	knownPlanets = new HashMap();
	//        name,      mass,     radius,   gravity, period
	addPlanet("Mercury", 3.30e23,  2439e3,   3.7,     5.067023e6);
	addPlanet("Venus",   4.84e24,  6051.5e3, 8.865,   -20.99e6);
        addPlanet("Earth",   5.976e24, 6376.7e3, 9.80665, 86164.09);
	addPlanet("Luna",    7.352e22, 1738e3,   1.624,   2.3606e6);
	addPlanet("Mars",    6.418e23, 3392.9e3, 3.72,    88642.66);
	addPlanet("Jupiter", 1.899e27, 69824e3,  25.99,   50430);
	addPlanet("Saturn",  5.693e26, 120660e3, 2.60909, 38364);
	/* Britanica does not give the rotational period for Jupiter's moons...
	addPlanet("Io",     8.886e22,  1815e3,   ,        );
	addPlanet("Europa", 4.785e22,  1569e3,   ,        );
	addPlanet("Ganymede",1.481e23, 2631e3,   ,        );
	addPlanet("Callisto",1.075e23, 2400e3,   ,        );
	 */

	//Values from Britannica for Mercury, Venus, Earth rotation, Luna,
	//Mars, Deimos, Phobos, Jupiter, Juptier's mooons, Saturn.

	//Regarding the values for Earth:
	  //Period = 86164.09 = 60*60*24.0*(1.0-1.0/366.2414) seconds
	    //One day on Earth is a bit more than one revolution, since the
	    //Earth is traveling around the sun at the same time; thus one
	    //revolution is a little bit less than one day, by about the
	    //amount that the Earth travels around the sun in that one day.
	    //The divisor is 366.xxx rather than 365.xxx to make up for the
	    //day "lost" in a year because of one trip around the sun.
	    //The fractional part (.2414) is the "leap year" partial day length.
	  //Gravity = 9.80665
	    //This value is from Mathematica (off the web)
	  //Radius = 6376700.0
	    //This radius value was derived from the Mathematica values for
	    //the Earth's surface gravity (above) and mass (5.976e24 kg).
	    //The IUGG value for earth radius is 6378137 (6378km).
	    //Britannica gives a polar radius of 6356784 and an equatorial
	    //radius of 6378136.
	  //About the value for GEO altitude:
	    //The flattening of the earth is 0.00335281, i.e. the polar radius
	    //is (1-f) times the equatorial radius, so the radius of the earth
	    //is about .0033 * 6300km larger at the equator, or about 20km.
	    //Our model assumes a perfectly spherical Earth; we don't get
	    //exactly the right value for GEO altitude, we are off by a few km
	    //(GEO is 35,786 km, but we calculate it as 35,793 km).

	//Mars: Radius = 3396.9e3 is the equatorial radius, polar is 3376.1e3
	//  the value 3392.9e3 gives the correct mass.
	//Mars' satellites Deimos and Phobos are pretty much just rocks,
	//each with mass under 1e16 kg and average radius in the neighborhood
	//of 10 km.  This puts their surface gravity at under 0.01 m/s^2,
	//and escape velocity at 4.7 m/s for Deimos and 7.8 m/s for Phobos,
	//which means you could almost reach escape velocity with a good jump.

	//Jupiter: Radius = 71400e3 is the equatorial radius, polar is 66770e3.
	//A radius of 69824e3 gives the right mass.
	//Period = 50430 is the equatorial period.  The period in the higher
	//latitudes is 35740, and the period of the magnetic field is 35729.
	//Britannica claims surface gravity on Jupiter is 22.88, which they
	//say is 2.65 times Earth; clearly not, since 2.65*9.8=26, quite far
	//from 22.88.  So I just used 2.65*9.80665=25.99.

	//Saturn: Radius = 120660e3 equatorial, 108000e3 polar.
    }

    private static void addPlanet(String name, double mass, double radius,
    		double gravity, double period) {
        Planet p = new Planet();
	p.setMass(mass);
	p.setRadius(radius);
	//p.setGravity(gravity);
		//let surface gravity be calculated based on mass and radius
	p.setPeriod(period);
	knownPlanets.put(name,p);
    }

    /** Create a planet object.
     */
    public Planet() {}

    /** Set the mass of the planet.
     * @param mass The mass of the planet in kg.
     */
    public void setMass(double mass) {
    	this.mass = mass;
    }

    /** Get the mass of the planet.
     * @return The mass of the planet in kg.
     */
    public double getMass() {
    	return mass;
    }

    /** Set the radius of the planet.
     * @param radius The radius of the planet in meters.
     */
    public void setRadius(double radius) {
    	this.radius = radius;
    }

    /** Get the radius of the planet.
     * @return The radius of the planet in meters.
     */
    public double getRadius() {
    	return radius;
    }

    /** Set the rotation period of the planet.
     * @param period The rotation period of the planet
     *        in meters per second-squared.
     */
    public void setPeriod(double period) {
    	this.period = period;
    }

    /** Get the rotation period of the planet.
     * @return The rotation period of the planet in seconds.
     */
    public double getPeriod() {
    	return period;
    }

    /** Get the surface gravity of the planet.
     * @return The surface gravity of the planet in meters per second-squared.
     */
    public double getGravity() {
	double r = getRadius();
	if (r==0.0)
	    return 0.0;		//no radius set
	double m = getMass();
	return mass*Physics.G/(r*r);
    }

    /** Get the density of the planet based on its radius and mass.
     */
    public double getDensity() {
        double r = getRadius();
	if (r==0.0)
	    return 0.0;		//no radius set
        double mass = getMass();
	double volume = (4.0/3.0)*Math.PI*(r*r*r);
	return mass/volume;
    }

    /** Get the altitude of synchronous orbit.
     * @return The height of synchronious orbit above the surface in meters.
     */
    public double getSynchronousAltitude() {
	double g = getGravity();
    	double r = getRadius();
	double p = getPeriod();
	double h3 = Physics.G*getMass()*p*p/(4*Math.PI*Math.PI);
	double h = Math.pow(h3,(1.0/3.0));	//cube root
		//h = height of sync orbit from center of planet
	return h - r;
    }

    /** Get the current rotation angle of the prime meridian.
     * The rotation at time zero is defined to be zero.
     * @return The counterclockwise rotation in radians.
     */
    public double getPrimeAngle() {
    	return primeAngle;
    }

    /** Reset the simulation. */
    protected void resetSimulation() {
        primeAngle = 0;
    }

    /** Simulate a time change of the specified amount. */
    protected void simulateOneStep(double simInterval) {
        primeAngle += (simInterval/period)*2*Math.PI;
        if (primeAngle>=2*Math.PI || -primeAngle>=2*Math.PI) {
	    primeAngle = Math.IEEEremainder(primeAngle,2*Math.PI);
	}
    }

    /** Get the position of the base station. */
    public DVector getBase() {
        return getBase(0.0);
    }

    /** Get a future position of the base station. */
    public DVector getBase(double simOffset) {
        double offsetAngle = (simOffset/period)*2*Math.PI;
        return new DVector(radius*Math.cos(primeAngle+offsetAngle),
                radius*Math.sin(primeAngle+offsetAngle));
    }

    /** Get the current velocity of the base station. */
    public DVector getBaseVelocity() {
        return getBaseVelocity(0.0);
    }

    /** Get a future velocity of the base station. */
    public DVector getBaseVelocity(double simOffset) {
        double baseAngle = primeAngle+(simOffset/period)*2*Math.PI;
        double speed = 2*Math.PI*radius/period;
        return DVector.newPolar(speed,primeAngle+Math.PI/2.0);
    }

    /** Get the velocity of the surface of the planet at the given point.
     * We ignore the radius and just take the angle of the point, then
     * return the velocity of the planet surface at that angle.
     */
    public DVector getGroundVelocity(DVector g) {
        double angle = g.getAngle();
        double speed = 2*Math.PI*radius/period;
        return DVector.newPolar(speed,angle+Math.PI/2.0);
    }

    /** Get the theta coordiate of the base station. */
    public double getBaseTheta() {
        return primeAngle;
    }

    /** Draw our planet at the specified scale, and time. */
    public void draw(Graphics2D g, double scale) {
        Color oldColor = g.getColor();
        g.setColor(planetColor);
    	int r = (int)(scale*getRadius());
    	g.drawArc(-r,-r,2*r,2*r,0,360);
	double a = primeAngle;
	int rx = (int)(r*Math.cos(a));
	int ry = (int)(r*Math.sin(a));
	g.drawLine(0,0,rx,ry);
	    //draw a line indicating the prime meridian
        g.setColor(oldColor);
    }
}

//end
