/* SpelCanvas.java
 *
 * Jim McBeath, March 14, 2003
 */

package net.jimmc.spelsim;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Panel;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class SpelCanvas extends Panel implements Runnable {

    private long lastDrawTime;      //The time at which we last drew our image
    private long lastTickTime;
    private SimEvent[] simEvents;

    /** The sim to which we are attached. */
    SpelSim sim;

    /** Our sim listeners. */
    List simListenerList = new ArrayList();

    /** The current simulation ticks per second. */
    double ticksPerSecond;

    /** The tick interval, inverse of ticks per second. */
    double tickInterval;

    /** The amount of simulated time represented by one tick. */
    double simInterval;

    /** The damping factor. */
    double dampingFactor;

    /** The counterweight derating.  We reduce the effective mass of the
     * counterweight by this much in order to compensate for some of
     * the high-frequency artifacts that arise due to our discrete
     * element simulation. */
    double cwDerating;

    /** Drag when an element is on the ground. */
    double groundDrag;

    /** The maximum number of screen updates to do per second. */
    double maxRefreshRate;
    private int minDrawInterval;        //minimum millis between updates

    /** True to display using the planet's perspective, where the
     * universe rotates and the planet stays still.
     */
    boolean planetPerspective;

    //True to use a real bounding box so that everything is in the view,
    //false to stick with the initial fixed view.
    int viewScale;
        public final static int VIEW_SCALE_FIXED = 0;
        public final static int VIEW_SCALE_BOUNDING = 1;
        public final static int VIEW_SCALE_PLANET = 2;

    /** The step integration algorithm to use. */
    int algorithm;
        public final static int ALG_FORWARD = 0;
        public final static int ALG_FORWARD_CIRCULAR = 1;
        public final static int ALG_TRAPEZOID = 2;
        public final static int MAX_ALG_VALUE = ALG_TRAPEZOID;

    /** True to enable debug output. */
    boolean debugOutputEnabled;

    /** True if our simulation is running. */
    boolean running;

    /** The current simulation time. */
    double time;

    /** The simulation thread. */
    Thread simThread;

    /** True if a reset has been requested. */
    private volatile boolean needReset;

    public SpelCanvas(SpelSim sim) {
    	this.sim = sim;
	setBackground(Color.white);
	setForeground(Color.black);
        setSimStepSize(1);      //put in the default simulation parameters
	setTicksPerSecond(1000);
        setDampingFactor(0.1);
        setCwDerating(4e-4);
        setGroundDrag(1.0);
        setMaxRefreshRate(20);
	time = 0.0;
    }

    public Dimension getPreferredSize() {
	return new Dimension(400,400);
    }

    /** Set the size of each simulation step in sim-seconds. */
    public void setSimStepSize(double simSeconds) {
        if (simSeconds<0.0)     //negative numbers not allowed
            simSeconds = 1.0;
        simInterval = simSeconds;
        if (simThread!=null)
	    simThread.interrupt();	//make it use the new interval
    }

    public double getSimStepSize() {
        return simInterval;
    }

    /** Set the number of simulation ticks we perform per second. */
    public void setTicksPerSecond(double d) {
	ticksPerSecond = d;
	if (ticksPerSecond==0.0)
	    tickInterval = 0;	//undefined
	else
	    tickInterval = 1.0/d;	//sleep time between ticks
        if (simThread!=null)
	    simThread.interrupt();	//make it use the new interval
    }

    /** Get the number of simulation ticks per second. */
    public double getTicksPerSecond() {
    	return ticksPerSecond;
    }

    /** Get the simulation speed.
     * This is the ratio of simulation time to real time.
     */
    public double getSimulationSpeed() {
    	return simInterval / tickInterval;
    }

    /** Set the damping factor to use. */
    public void setDampingFactor(double df) {
        dampingFactor = df;
    }

    /** Get the damping factor. */
    public double getDampingFactor() {
        return dampingFactor;
    }

    /** Set the CW derating to use. */
    public void setCwDerating(double df) {
        cwDerating = df;
    }

    /** Get the CW derating. */
    public double getCwDerating() {
        return cwDerating;
    }

    /** Set the ground drag to use. */
    public void setGroundDrag(double drag) {
        groundDrag = drag;
    }

    /** Get the ground drag. */
    public double getGroundDrag() {
        return groundDrag;
    }

    /** Set the maximum refresh rate.
     * We will not update the screen any more often than
     * this many times per second.
     */
    public void setMaxRefreshRate(double maxPerSecond) {
        if (maxPerSecond<=0.0)  //don't allow negative values
            maxPerSecond = 1.0;
        maxRefreshRate = maxPerSecond;
        minDrawInterval = (int)(1000/maxRefreshRate);
    }

    /** Get the maximum refresh rate.
     */
    public double getMaxRefreshRate() {
        return maxRefreshRate;
    }

    /** Set the perspective to use.
     * @param perspective True to use the planet's perspective, false to
     *        use sidereal perspective.
     */
    public void setPerspective(boolean perspective) {
        this.planetPerspective = perspective;
    }

    /** Set the scale/view to use.
     * @param viewScale One of the VIEW_SCALE_* constants.
     */
    public void setViewScale(int viewScale) {
        this.viewScale = viewScale;
        repaint();
    }

    /** Set the step integration algorithm to use.
     * @param alg One of the ALG_* constants.
     */
    public void setAlgorithm(int alg) {
        if (alg<0 || alg>MAX_ALG_VALUE)
            throw new IllegalArgumentException("Illegal algorithm value");
        this.algorithm = alg;
    }

    public int getAlgorithm() {
        return this.algorithm;
    }

    /** Set the debug flag.
     * @param enabled True to print debug output.
     */
    public void setDebugOutput(boolean enabled) {
        this.debugOutputEnabled = enabled;
        getCable().setDebugOutputEnabled(enabled);
    }

    /** Set the simulation events. */
    public void setSimEvents(SimEvent[] simEvents) {
        if (sameEvents(simEvents,this.simEvents))
            return;             //no change
        //If we are already running, assume every event which is in
        //the past has already happened.
        double currentTime = time;
        for (int i=0; i<simEvents.length; i++) {
            if (simEvents[i].getTime()<currentTime)
                simEvents[i].setDone();
        }
        this.simEvents = simEvents;
    }

    private boolean sameEvents(SimEvent[] a, SimEvent[] b) {
        if (a==b)
            return true;
        if (a==null || b==null)
            return false;
        if (a.length!=b.length)
            return false;
        for (int i=0; i<a.length; i++) {
            if (!a[i].equals(b[i]))
                return false;
        }
        return true;    //all events match
    }

    public void addSimListener(ISimListener listener) {
        simListenerList.add(listener);
    }

    public void start() {
	if (running)
	    return;		//already running
    	//TODO - make sure old thread is really done
	simThread = new Thread(this);
	simThread.setPriority(simThread.getPriority()-1);
	running = true;
	simThread.start();
    }

    public void stop() {
	if (!running || simThread==null)
	    return;	//not running, do nothing
    	running = false;
	simThread.interrupt();		//wake it up so it can stop
    }

    public void step() {
        if (running || simThread!=null) {
	    System.out.println("Already running");
	    return;	//don't try to step when we are running
	}
	simulateOneStep();
    }

    /** Reset the simulation. */
    public void reset() {
        if (running)
	    needReset = true;
	else
	    doReset();
    }

    /** Do the actual reset.  Don't call this from another thread while
     * the simulation is running, as it can cause a Null Pointer Exception
     * when the cable elements are being recalculated.
     */
    private void doReset() {
        time = 0.0;
	sim.getPlanet().resetSimulation();
	sim.getCable().resetSimulation();
        resetEvents();
	needReset = false;
	repaint();
        if (debugOutputEnabled)
            sim.getCable().dumpSelectedElements(time,getAlgorithm());

        //Invoke all of our listeners
        Iterator it = simListenerList.iterator();
        while (it.hasNext()) {
            ISimListener simListener = (ISimListener)it.next();
            simListener.simStepped(this,true);
        }
    }

    private void resetEvents() {
        if (simEvents==null)
            return;
        for (int i=0; i<simEvents.length; i++) {
            simEvents[i].setDone(false);
        }
    }

    public void run() {
    	while (running) {
	    if (needReset) {
	        doReset();
	    }
	    simulateOneStep();
	    waitForTick();
	}
	repaint();
	if (Thread.currentThread()==simThread)
	    simThread = null;
    }

    /** Simulate one clock tick. */
    protected void simulateOneStep() {
        simulateEvents();
	sim.getCable().simulateOneStep(simInterval,getAlgorithm(),
                dampingFactor,cwDerating, groundDrag);
	sim.getPlanet().simulateOneStep(simInterval);
	time += simInterval;
        long now = System.currentTimeMillis();
        boolean didDraw = false;
        if (now - lastDrawTime > minDrawInterval) {
            //Only redraw if enough time has passed
            lastDrawTime = now;
            repaint();	//redraw our image
            didDraw = true;
        }
        if (debugOutputEnabled)
            sim.getCable().dumpSelectedElements(time,getAlgorithm());

        //Invoke all of our listeners
        Iterator it = simListenerList.iterator();
        while (it.hasNext()) {
            ISimListener simListener = (ISimListener)it.next();
            simListener.simStepped(this,didDraw);
        }
    }

    protected void simulateEvents() {
        if (simEvents==null)
            return;             //no events
        for (int i=0; i<simEvents.length; i++) {
            SimEvent ev = simEvents[i];
            if (ev.isDone())
                continue;
            if (ev.getTime()>time)
                break;          //haven't gotten there yet
            simulateEvent(ev);  //simulate this event now
        }
    }

    private void simulateEvent(SimEvent ev) {
//System.out.println("Simulating event at time "+ev.getTime());
        switch (ev.getType()) {
        case SimEvent.TYPE_BREAK:
            int elementNumber = ev.getArg0Int();
            getCable().setElementBroken(elementNumber);
            break;
        case SimEvent.TYPE_NOP:
        default:
            //do nothing
            break;
        }
        ev.setDone();
    }

    /** Wait for the next simulation tick. */
    protected void waitForTick() {
        long now = System.currentTimeMillis();
        if (lastTickTime==0)
            lastTickTime = now; //oversleep on first tick rather than no sleep
        long sinceLastTick = now - lastTickTime;
        long waitTime = (long)(tickInterval*1000) - sinceLastTick;
        if (waitTime>0) {
            try {
                Thread.sleep((int)waitTime);
            } catch (InterruptedException ex) {
                //awakened from our sleep
            }
        }
        lastTickTime = now + waitTime;
    }

    public void paint(Graphics graphics) {
        //Do our scale calculations
    	Cable cable = getCable();
	Planet planet = getPlanet();
        Rectangle2D bbox = getBoundingBox();
	int panelHeight = getHeight();
	int panelWidth = getWidth();
        double xScale = ((double)panelWidth)/bbox.getWidth();
        double yScale = ((double)panelHeight)/bbox.getHeight();
        double scale = (xScale<yScale)?xScale:yScale;
        double xTran = (-bbox.getX()/bbox.getWidth())*scale/xScale +
                (1.0 - scale/xScale)/2.0;
        double yTran = (-bbox.getY()/bbox.getHeight())*scale/yScale +
                (1.0 - scale/yScale)/2.0;

        //Make the changes to our graphics context
	Graphics2D g = (Graphics2D)graphics.create();
	//drawTime(g);
        //drawBrokenElementCount(g);
        drawScale(g,scale);
	g.scale(1.0,-1.0);	//turn it upside down
	g.translate(0,-getHeight());
                    // (i.e. use 1st quadrant coordinates)
	g.translate((double)panelWidth*xTran, (double)panelHeight*yTran);
		//Planet is typically in the middle of the drawing area,
                //but may not be when using Bounding view scale.
	if (planetPerspective) {
	    g.rotate(-planet.getPrimeAngle());
	}

        //Draw the planet and cable
	planet.draw(g,scale);		//draw the planet
	cable.draw(g,scale);		//draw the cable
    }

    private Rectangle2D getBoundingBox() {
        Rectangle2D bbox;
        switch (viewScale) {
        default:
        case VIEW_SCALE_FIXED:
            bbox = getFixedBoundingBox();
            break;
        case VIEW_SCALE_BOUNDING:
            bbox = getActualBoundingBox();
            break;
        case VIEW_SCALE_PLANET:
            bbox = getPlanetBoundingBox();
            break;
        }
        //Add a small margin on each edge
        double margin = 0.03;
        double width = bbox.getWidth();
        double height = bbox.getHeight();
        bbox.add(bbox.getX() - margin*width, bbox.getY() - margin*height);
        bbox.add(bbox.getX()+(width*(1.0+2.0*margin)),
                bbox.getY()+(height*(1.0+2.0*margin)));
        return bbox;
    }

    //Get a bounding box that contains the planet and all cable elements
    private Rectangle2D getActualBoundingBox() {
        if (planetPerspective) {
            //If planet perspective, then the planet has to go at the
            //center as we rotate everything else around it.
            //So we need the max radius of any cable element.
            Planet planet = getPlanet();
            Cable cable = getCable();
            double r = planet.getRadius();
            double cr = cable.getBoundingRadius();
            if (cr>r)
                r = cr;
            return new Rectangle2D.Double(-r, -r, 2.0*r, 2.0*r);
        }

        Rectangle2D bbox = getFixedBoundingBox();
            //start with a minimum of the fixed size, which includes the planet

        //Ensure that every cable element is in the picture
    	Cable cable = getCable();
        bbox.add(cable.getBoundingBox());

        return bbox;
    }

    //Get a bounding box that contains the planet and cable in its initial
    //orbit.
    private Rectangle2D getFixedBoundingBox() {
    	Cable cable = getCable();
	Planet planet = getPlanet();
	double radius = planet.getRadius();
	double length = radius*2 + cable.getLength()*2;
		//big enough to draw the planet in the middle and the
		//cable sticking out in any direction
        double w = length;
        double h = length;
        double x = -length/2.0;
        double y = -length/2.0;
        Rectangle2D bbox = new Rectangle2D.Double(x,y,w,h);
        return bbox;
    }

    //Get a bounding box that contains only the planet and a bit more.
    private Rectangle2D getPlanetBoundingBox() {
	Planet planet = getPlanet();
        double r = planet.getRadius();
        r *= 1.1;       //get a bit more
        Rectangle2D bbox = new Rectangle2D.Double(-r,-r,2.0*r,2.0*r);
        return bbox;
    }

    /** Draw the current scale into the window.
     */
    protected void drawScale(Graphics2D g, double scale) {
	int x = 10;
	int y = getHeight()-15;
        //TODO - draw a scale line; format scale number shorter
        //g.drawLine(x,y,x+100,y);
        String s = "Scale: "+scale;
        g.drawString(s,x,y);
    }

    /** Draw the current time value into the window.
     * @param g The unmodified Graphics for our drawing window
     */
    protected void drawTime(Graphics2D g) {
	String r = running?"Running":"Stopped";
    	String s = "Time="+time;
	int x = 10;
	int y = getHeight()-40;
	g.drawString(r,x,y);
	y += 20;
	g.drawString(s,x,y);
    }

    public String getRunStatus() {
	return running?"Running":"Stopped";
    }
    public double getTime() {
        return time;
    }
    public String getBrokenElementCountString() {
    	Cable cable = getCable();
        return cable.getNumBrokenElements()+"/"+cable.getNumElements();
    }
    public String getGroundedElementCountString() {
    	Cable cable = getCable();
        return cable.getNumGroundedElements()+"/"+cable.getNumElements();
    }
    public String getStatusString() {
        StringBuffer sb = new StringBuffer();
        sb.append(getRunStatus()).append("; ");
        sb.append("Time=").append(time).append("\n");
        Planet planet = getPlanet();
        sb.append("Base angle=").append(planet.getPrimeAngle()).append("\n");
        sb.append("Broken=").append(getBrokenElementCountString()).append("\n");
        sb.append("Grounded=").append(
                getGroundedElementCountString()).append("\n");
        return sb.toString();
    }

    protected void drawBrokenElementCount(Graphics2D g) {
        String s = getBrokenElementCountString();
        int x = getWidth() - 200;
        int y = getHeight() - 20;
        g.drawString(s,x,y);
    }

    public Cable getCable() {
    	return sim.getCable();
    }

    public Planet getPlanet() {
    	return sim.getPlanet();
    }
}

//end
